from __future__ import annotations

import datetime
import logging
from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum
from typing import FrozenSet, List, Optional, Sequence, Tuple

from dbt_semantic_interfaces.implementations.elements.dimension import PydanticDimensionTypeParams
from dbt_semantic_interfaces.implementations.filters.where_filter import PydanticWhereFilter
from dbt_semantic_interfaces.references import EntityReference, MeasureReference, MetricReference
from dbt_semantic_interfaces.type_enums import DimensionType
from metricflow_semantics.dag.sequential_id import SequentialIdGenerator
from metricflow_semantics.errors.error_classes import ExecutionException
from metricflow_semantics.filters.time_constraint import TimeRangeConstraint
from metricflow_semantics.mf_logging.lazy_formattable import LazyFormat
from metricflow_semantics.mf_logging.runtime import log_block_runtime
from metricflow_semantics.model.linkable_element_property import LinkableElementProperty
from metricflow_semantics.model.semantic_manifest_lookup import SemanticManifestLookup
from metricflow_semantics.model.semantics.element_filter import LinkableElementFilter
from metricflow_semantics.model.semantics.linkable_element import LinkableDimension
from metricflow_semantics.model.semantics.semantic_model_helper import SemanticModelHelper
from metricflow_semantics.naming.linkable_spec_name import StructuredLinkableSpecName
from metricflow_semantics.protocols.query_parameter import GroupByParameter, MetricQueryParameter, OrderByQueryParameter
from metricflow_semantics.query.query_exceptions import InvalidQueryException
from metricflow_semantics.query.query_parser import MetricFlowQueryParser
from metricflow_semantics.random_id import random_id
from metricflow_semantics.specs.column_assoc import ColumnAssociationResolver
from metricflow_semantics.specs.dunder_column_association_resolver import DunderColumnAssociationResolver
from metricflow_semantics.specs.query_param_implementations import SavedQueryParameter
from metricflow_semantics.specs.query_spec import MetricFlowQuerySpec
from metricflow_semantics.specs.spec_set import InstanceSpecSet
from metricflow_semantics.sql.sql_table import SqlTable
from metricflow_semantics.time.time_source import TimeSource
from metricflow_semantics.time.time_spine_source import TimeSpineSource

from metricflow.data_table.mf_table import MetricFlowDataTable
from metricflow.dataflow.builder.builder_cache import DataflowPlanBuilderCache
from metricflow.dataflow.builder.dataflow_plan_builder import DataflowPlanBuilder
from metricflow.dataflow.builder.node_data_set import DataflowPlanNodeOutputDataSetResolver
from metricflow.dataflow.builder.source_node import SourceNodeBuilder
from metricflow.dataflow.dataflow_plan import DataflowPlan
from metricflow.dataflow.optimizer.dataflow_optimizer_factory import DataflowPlanOptimization
from metricflow.dataset.convert_semantic_model import SemanticModelToDataSetConverter
from metricflow.dataset.dataset_classes import DataSet
from metricflow.dataset.semantic_model_adapter import SemanticModelDataSet
from metricflow.engine.models import Dimension, Entity, Measure, Metric, SavedQuery
from metricflow.engine.time_source import ServerTimeSource
from metricflow.execution.convert_to_execution_plan import ConvertToExecutionPlanResult
from metricflow.execution.dataflow_to_execution import (
    DataflowToExecutionPlanConverter,
)
from metricflow.execution.execution_plan import ExecutionPlan, SqlStatement
from metricflow.execution.executor import SequentialPlanExecutor
from metricflow.plan_conversion.dataflow_to_sql import DataflowToSqlPlanConverter
from metricflow.protocols.sql_client import SqlClient
from metricflow.sql.optimizer.optimization_levels import SqlOptimizationLevel
from metricflow.telemetry.models import TelemetryLevel
from metricflow.telemetry.reporter import TelemetryReporter, log_call

logger = logging.getLogger(__name__)
_telemetry_reporter = TelemetryReporter(report_levels_higher_or_equal_to=TelemetryLevel.USAGE)
_telemetry_reporter.add_python_log_handler()


@dataclass(frozen=True)
class MetricFlowRequestId:
    """Uniquely identifies a request to the MF engine."""

    mf_rid: str


class MetricFlowQueryType(Enum):
    """An enum to designate what type of MetricFlow query it is."""

    METRIC = "metric"
    DIMENSION_VALUES = "dimension_values"


@dataclass(frozen=True)
class MetricFlowQueryRequest:
    """Encapsulates the parameters for a metric query.

    TODO: This has turned into a bag of parameters that make it difficult to use without a bunch of conditionals.

    metric_names: Names of the metrics to query.
    metrics: Metric objects to query.
    group_by_names: Names of the dimensions and entities to query.
    group_by: Dimension or entity objects to query.
    limit: Limit the result to this many rows.
    time_constraint_start: Get data for the start of this time range.
    time_constraint_end: Get data for the end of this time range.
    where_constraints: A sequence of SQL strings that can be used like a where clause on the output data.
    order_by_names: metric and group by names to order by. A "-" can be used to specify reverse order e.g. "-ds".
    order_by: metric, dimension, or entity objects to order by.
    output_table: If specified, output the result data to this table instead of a result data_table.
    sql_optimization_level: The level of optimization for the generated SQL.
    query_type: Type of MetricFlow query.
    """

    request_id: MetricFlowRequestId
    saved_query_name: Optional[str]
    metric_names: Optional[Sequence[str]]
    metrics: Optional[Sequence[MetricQueryParameter]]
    group_by_names: Optional[Sequence[str]]
    group_by: Optional[Tuple[GroupByParameter, ...]]
    limit: Optional[int]
    time_constraint_start: Optional[datetime.datetime]
    time_constraint_end: Optional[datetime.datetime]
    where_constraints: Optional[Sequence[str]]
    order_by_names: Optional[Sequence[str]]
    order_by: Optional[Sequence[OrderByQueryParameter]]
    min_max_only: bool
    sql_optimization_level: SqlOptimizationLevel
    dataflow_plan_optimizations: FrozenSet[DataflowPlanOptimization]
    query_type: MetricFlowQueryType

    @staticmethod
    def create_with_random_request_id(  # noqa: D102
        saved_query_name: Optional[str] = None,
        metric_names: Optional[Sequence[str]] = None,
        metrics: Optional[Sequence[MetricQueryParameter]] = None,
        group_by_names: Optional[Sequence[str]] = None,
        group_by: Optional[Tuple[GroupByParameter, ...]] = None,
        limit: Optional[int] = None,
        time_constraint_start: Optional[datetime.datetime] = None,
        time_constraint_end: Optional[datetime.datetime] = None,
        where_constraints: Optional[Sequence[str]] = None,
        order_by_names: Optional[Sequence[str]] = None,
        order_by: Optional[Sequence[OrderByQueryParameter]] = None,
        sql_optimization_level: SqlOptimizationLevel = SqlOptimizationLevel.default_level(),
        dataflow_plan_optimizations: FrozenSet[
            DataflowPlanOptimization
        ] = DataflowPlanOptimization.enabled_optimizations(),
        query_type: MetricFlowQueryType = MetricFlowQueryType.METRIC,
        min_max_only: bool = False,
    ) -> MetricFlowQueryRequest:
        return MetricFlowQueryRequest(
            request_id=MetricFlowRequestId(mf_rid=f"{random_id()}"),
            saved_query_name=saved_query_name,
            metric_names=metric_names,
            metrics=metrics,
            group_by_names=group_by_names,
            group_by=group_by,
            limit=limit,
            time_constraint_start=time_constraint_start,
            time_constraint_end=time_constraint_end,
            where_constraints=where_constraints,
            order_by_names=order_by_names,
            order_by=order_by,
            sql_optimization_level=sql_optimization_level,
            dataflow_plan_optimizations=dataflow_plan_optimizations,
            query_type=query_type,
            min_max_only=min_max_only,
        )


@dataclass(frozen=True)
class MetricFlowQueryResult:
    """The result of a query and context on how it was generated."""

    query_spec: MetricFlowQuerySpec
    dataflow_plan: DataflowPlan
    sql: str
    result_df: Optional[MetricFlowDataTable] = None
    result_table: Optional[SqlTable] = None


@dataclass(frozen=True)
class MetricFlowExplainResult:
    """Returns plans for resolving a query."""

    query_spec: MetricFlowQuerySpec
    dataflow_plan: DataflowPlan
    convert_to_execution_plan_result: ConvertToExecutionPlanResult
    output_table: Optional[SqlTable] = None

    @property
    def sql_statement(self) -> SqlStatement:
        """Return the SQL query that would be run for the given query."""
        execution_plan = self.execution_plan
        if len(execution_plan.tasks) != 1:
            raise NotImplementedError(
                str(
                    LazyFormat(
                        "Multiple tasks in the execution plan not yet supported.",
                        tasks=[task.task_id for task in execution_plan.tasks],
                    )
                )
            )

        sql_statement = execution_plan.tasks[0].sql_statement
        if not sql_statement:
            raise NotImplementedError(
                str(
                    LazyFormat(
                        "Execution plan tasks without a SQL statement are not yet supported.",
                        tasks=[task.task_id for task in execution_plan.tasks],
                    )
                )
            )

        return sql_statement

    @property
    def execution_plan(self) -> ExecutionPlan:  # noqa: D102
        return self.convert_to_execution_plan_result.execution_plan


class AbstractMetricFlowEngine(ABC):
    """Query interface for clients."""

    @abstractmethod
    def query(
        self,
        mf_request: MetricFlowQueryRequest,
    ) -> MetricFlowQueryResult:
        """Query for metrics."""
        pass

    @abstractmethod
    def explain(
        self,
        mf_request: MetricFlowQueryRequest,
    ) -> MetricFlowExplainResult:
        """Similar to query - returns the query that would have been executed."""
        pass

    @abstractmethod
    def simple_dimensions_for_metrics(
        self, metric_names: List[str], without_any_property: Sequence[LinkableElementProperty]
    ) -> List[Dimension]:
        """Retrieves a list of all common dimensions for metric_names.

        "simple" dimensions are the ones that people expect from a UI perspective. For example, if "ds" is a time
        dimension at a day granularity, this would not list "ds__week".

        Args:
            metric_names: Names of metrics to get common dimensions from.
            without_any_property: Return dimensions not matching these properties.

        Returns:
            A list of Dimension objects containing metadata.
        """
        pass

    @abstractmethod
    def entities_for_metrics(self, metric_names: List[str]) -> List[Entity]:
        """Retrieves a list of all entities for metric_names.

        Args:
            metric_names: Names of metrics to get common entities from.

        Returns:
            A list of Entity objects containing metadata.
        """
        pass

    @abstractmethod
    def list_metrics(self) -> List[Metric]:
        """Retrieves a list of metric names.

        Returns:
            A list of Metric objects containing metadata.
        """
        pass

    @abstractmethod
    def get_dimension_values(
        self,
        metric_names: List[str],
        get_group_by_values: str,
        time_constraint_start: Optional[datetime.datetime] = None,
        time_constraint_end: Optional[datetime.datetime] = None,
    ) -> List[str]:
        """Retrieves a list of dimension values given a [metric_name, get_group_by_values].

        Args:
            metric_names: Names of metrics that contain the group_by.
            get_group_by_values: Name of group_by to get values from.
            time_constraint_start: Get data for the start of this time range.
            time_constraint_end: Get data for the end of this time range.

        Returns:
            A list of dimension values as string.
        """
        pass

    @abstractmethod
    def explain_get_dimension_values(
        self,
        metric_names: Optional[List[str]] = None,
        metrics: Optional[Sequence[MetricQueryParameter]] = None,
        get_group_by_values: Optional[str] = None,
        group_by: Optional[GroupByParameter] = None,
        time_constraint_start: Optional[datetime.datetime] = None,
        time_constraint_end: Optional[datetime.datetime] = None,
    ) -> MetricFlowExplainResult:
        """Returns the SQL query for get_dimension_values.

        Args:
            metric_names: Names of metrics that contain the group_by.
            metrics: Similar to `metric_names`, but specified via parameter objects.
            get_group_by_values: Name of group_by to get values from.
            group_by: Similar to `get_group_by_values`, but specified via parameter objects.
            time_constraint_start: Get data for the start of this time range.
            time_constraint_end: Get data for the end of this time range.

        Returns:
            An object with the rendered SQL and generated plans.
        """
        pass

    @abstractmethod
    def list_dimensions(self) -> List[Dimension]:
        """List all dimensions in the semantic manifest."""
        pass


class MetricFlowEngine(AbstractMetricFlowEngine):
    """Main entry point for queries.

    Attributes on this class should be treated as in use by our APIs.
    TODO: provide a more stable API layer instead of assuming this class is stable.
    """

    # When generating IDs in the initializer, start from this value.
    _ID_ENUMERATION_START_VALUE_FOR_INITIALIZER = 10000
    # When generating IDs in queries, start from this value.
    _ID_ENUMERATION_START_VALUE_FOR_QUERIES = 0

    def __init__(
        self,
        semantic_manifest_lookup: SemanticManifestLookup,
        sql_client: SqlClient,
        time_source: TimeSource = ServerTimeSource(),
        query_parser: Optional[MetricFlowQueryParser] = None,
        column_association_resolver: Optional[ColumnAssociationResolver] = None,
        consistent_id_enumeration: Optional[bool] = True,
    ) -> None:
        """Initializer for MetricFlowEngine.

        consistent_id_enumeration can be set to True to reset the numbering of sequentially generated IDs on each query. This
        will help generate consistent SQL between queries as aliases will be the same.

        For direct calls to construct MetricFlowEngine, do not pass the following parameters,
        - time_source
        - column_association_resolver
        - time_spine_source

        These parameters are mainly there to be overridden during tests.
        """
        self._reset_id_enumeration = consistent_id_enumeration
        if self._reset_id_enumeration:
            # Some of the objects that are created below use generated IDs. To avoid collision with IDs that are
            # generated for queries, set the ID generation numbering to start at a high enough number.
            logger.debug(
                LazyFormat(
                    lambda: f"For creating setup objects, setting numbering of generated IDs to start at: "
                    f"{MetricFlowEngine._ID_ENUMERATION_START_VALUE_FOR_INITIALIZER}"
                )
            )
            SequentialIdGenerator.reset(MetricFlowEngine._ID_ENUMERATION_START_VALUE_FOR_INITIALIZER)
        self._semantic_manifest_lookup = semantic_manifest_lookup
        self._sql_client = sql_client
        self._column_association_resolver = column_association_resolver or (DunderColumnAssociationResolver())
        self._time_source = time_source
        self._time_spine_sources = TimeSpineSource.build_standard_time_spine_sources(
            semantic_manifest_lookup.semantic_manifest
        )
        self._source_data_sets: List[SemanticModelDataSet] = []
        converter = SemanticModelToDataSetConverter(column_association_resolver=self._column_association_resolver)
        for semantic_model in sorted(
            self._semantic_manifest_lookup.semantic_manifest.semantic_models, key=lambda model: model.name
        ):
            data_set = converter.create_sql_source_data_set(semantic_model)
            self._source_data_sets.append(data_set)
            logger.debug(LazyFormat(lambda: f"Created source dataset from semantic model '{semantic_model.name}'"))

        source_node_builder = SourceNodeBuilder(
            column_association_resolver=self._column_association_resolver,
            semantic_manifest_lookup=self._semantic_manifest_lookup,
        )
        source_node_set = source_node_builder.create_from_data_sets(self._source_data_sets)

        node_output_resolver = DataflowPlanNodeOutputDataSetResolver(
            column_association_resolver=self._column_association_resolver,
            semantic_manifest_lookup=self._semantic_manifest_lookup,
        )
        node_output_resolver.cache_output_data_sets(source_node_set.all_nodes)

        self._dataflow_plan_builder_cache = DataflowPlanBuilderCache()
        self._dataflow_plan_builder = DataflowPlanBuilder(
            source_node_set=source_node_set,
            semantic_manifest_lookup=self._semantic_manifest_lookup,
            column_association_resolver=self._column_association_resolver,
            node_output_resolver=node_output_resolver,
            source_node_builder=source_node_builder,
            dataflow_plan_builder_cache=self._dataflow_plan_builder_cache,
        )
        self._to_sql_plan_converter = DataflowToSqlPlanConverter(
            column_association_resolver=self._column_association_resolver,
            semantic_manifest_lookup=self._semantic_manifest_lookup,
        )
        self._executor = SequentialPlanExecutor()

        self._query_parser = query_parser or MetricFlowQueryParser(
            semantic_manifest_lookup=self._semantic_manifest_lookup,
        )

    @log_call(module_name=__name__, telemetry_reporter=_telemetry_reporter)
    def query(self, mf_request: MetricFlowQueryRequest) -> MetricFlowQueryResult:  # noqa: D102
        logger.info(LazyFormat("Starting query request", mf_request=mf_request))
        explain_result = self._create_execution_plan(mf_request)
        execution_plan = explain_result.convert_to_execution_plan_result.execution_plan

        if len(execution_plan.tasks) != 1:
            raise NotImplementedError("Multiple tasks not yet supported.")

        task = execution_plan.tasks[0]

        logger.debug(LazyFormat(lambda: f"Sequentially running tasks in:\n" f"{execution_plan.structure_text()}"))
        execution_results = self._executor.execute_plan(execution_plan)
        logger.debug(LazyFormat(lambda: "Finished running tasks in execution plan"))

        if execution_results.contains_task_errors:
            raise ExecutionException(f"Got errors while executing tasks:\n{execution_results.get_result(task.task_id)}")

        task_execution_result = execution_results.get_result(task.task_id)

        assert task_execution_result.sql, "Task execution should have returned SQL that was run"

        logger.info(LazyFormat("Finished query request", request_id=mf_request.request_id))
        return MetricFlowQueryResult(
            query_spec=explain_result.query_spec,
            dataflow_plan=explain_result.dataflow_plan,
            sql=task_execution_result.sql,
            result_df=task_execution_result.df,
            result_table=explain_result.output_table,
        )

    @property
    def all_time_constraint(self) -> TimeRangeConstraint:
        """TimeRangeConstraint representing the min & max dates supported."""
        return TimeRangeConstraint.all_time()

    def _create_execution_plan(self, mf_query_request: MetricFlowQueryRequest) -> MetricFlowExplainResult:
        if self._reset_id_enumeration:
            logger.debug(
                LazyFormat(
                    lambda: f"Setting ID generation to start at: {MetricFlowEngine._ID_ENUMERATION_START_VALUE_FOR_QUERIES}"
                )
            )
            SequentialIdGenerator.reset(MetricFlowEngine._ID_ENUMERATION_START_VALUE_FOR_QUERIES)

        if mf_query_request.saved_query_name is not None:
            if mf_query_request.metrics or mf_query_request.metric_names:
                raise InvalidQueryException("Metrics can't be specified with a saved query.")
            if mf_query_request.group_by or mf_query_request.group_by_names:
                raise InvalidQueryException("Group by items can't be specified with a saved query.")
            query_spec = self._query_parser.parse_and_validate_saved_query(
                saved_query_parameter=SavedQueryParameter(mf_query_request.saved_query_name),
                where_filters=(
                    [
                        PydanticWhereFilter(where_sql_template=where_constraint)
                        for where_constraint in mf_query_request.where_constraints
                    ]
                    if mf_query_request.where_constraints is not None
                    else None
                ),
                limit=mf_query_request.limit,
                time_constraint_start=mf_query_request.time_constraint_start,
                time_constraint_end=mf_query_request.time_constraint_end,
                order_by_names=mf_query_request.order_by_names,
                order_by_parameters=mf_query_request.order_by,
            ).query_spec
        else:
            query_spec = self._query_parser.parse_and_validate_query(
                metric_names=mf_query_request.metric_names,
                metrics=mf_query_request.metrics,
                group_by_names=mf_query_request.group_by_names,
                group_by=mf_query_request.group_by,
                limit=mf_query_request.limit,
                time_constraint_start=mf_query_request.time_constraint_start,
                time_constraint_end=mf_query_request.time_constraint_end,
                where_constraint_strs=mf_query_request.where_constraints,
                order_by_names=mf_query_request.order_by_names,
                order_by=mf_query_request.order_by,
                min_max_only=mf_query_request.min_max_only,
            ).query_spec
        logger.debug(LazyFormat("Parsed query", query_spec=query_spec))

        output_selection_specs: Optional[InstanceSpecSet] = None
        if mf_query_request.query_type == MetricFlowQueryType.DIMENSION_VALUES:
            # Filter result by dimension columns if it's a dimension values query
            if len(query_spec.entity_specs) > 0:
                raise InvalidQueryException("Querying dimension values for entities is not allowed.")
            output_selection_specs = InstanceSpecSet(
                dimension_specs=query_spec.dimension_specs,
                time_dimension_specs=query_spec.time_dimension_specs,
            )
        if query_spec.metric_specs:
            logger.info(
                LazyFormat(
                    "Building dataflow plan", dataflow_plan_optimizations=mf_query_request.dataflow_plan_optimizations
                )
            )
            dataflow_plan = self._dataflow_plan_builder.build_plan(
                query_spec=query_spec,
                output_selection_specs=output_selection_specs,
                optimizations=mf_query_request.dataflow_plan_optimizations,
            )
        else:
            logger.info(
                LazyFormat(
                    "Building dataflow plan for distinct values",
                    dataflow_plan_optimizations=mf_query_request.dataflow_plan_optimizations,
                )
            )

            dataflow_plan = self._dataflow_plan_builder.build_plan_for_distinct_values(
                query_spec=query_spec, optimizations=mf_query_request.dataflow_plan_optimizations
            )

        if len(dataflow_plan.sink_nodes) > 1:
            raise NotImplementedError(
                f"Multiple output nodes in the dataflow plan not yet supported. "
                f"Got tasks: {dataflow_plan.sink_nodes}"
            )

        logger.info(LazyFormat("Building execution plan"))
        _to_execution_plan_converter = DataflowToExecutionPlanConverter(
            sql_plan_converter=self._to_sql_plan_converter,
            sql_plan_renderer=self._sql_client.sql_plan_renderer,
            sql_client=self._sql_client,
            sql_optimization_level=mf_query_request.sql_optimization_level,
        )
        convert_to_execution_plan_result = _to_execution_plan_converter.convert_to_execution_plan(dataflow_plan)
        return MetricFlowExplainResult(
            query_spec=query_spec,
            dataflow_plan=dataflow_plan,
            convert_to_execution_plan_result=convert_to_execution_plan_result,
        )

    @log_call(module_name=__name__, telemetry_reporter=_telemetry_reporter)
    def explain(self, mf_request: MetricFlowQueryRequest) -> MetricFlowExplainResult:  # noqa: D102
        with log_block_runtime("explain"):
            return self._create_execution_plan(mf_request)

    def get_measures_for_metrics(self, metric_names: List[str]) -> List[Measure]:  # noqa: D102
        metrics = self._semantic_manifest_lookup.metric_lookup.get_metrics(
            metric_references=[MetricReference(element_name=metric_name) for metric_name in metric_names]
        )
        semantic_model_lookup = self._semantic_manifest_lookup.semantic_model_lookup

        measures = set()
        for metric in metrics:
            for input_measure in metric.input_measures:
                measure_reference = MeasureReference(element_name=input_measure.name)
                # populate new obj
                measure = semantic_model_lookup.measure_lookup.get_measure(measure_reference=measure_reference)
                measures.add(
                    Measure(
                        name=measure.name,
                        agg=measure.agg,
                        agg_time_dimension=semantic_model_lookup.measure_lookup.get_properties(
                            measure_reference=measure_reference
                        ).agg_time_dimension_reference.element_name,
                        description=measure.description,
                        expr=measure.expr,
                        agg_params=measure.agg_params,
                    )
                )
        return list(measures)

    def simple_dimensions_for_metrics(  # noqa: D102
        self,
        metric_names: List[str],
        without_any_property: Sequence[LinkableElementProperty] = (
            LinkableElementProperty.ENTITY,
            LinkableElementProperty.DERIVED_TIME_GRANULARITY,
            LinkableElementProperty.DATE_PART,
            LinkableElementProperty.LOCAL_LINKED,
        ),
    ) -> List[Dimension]:
        path_key_to_linkable_dimensions = (
            self._semantic_manifest_lookup.metric_lookup.linkable_elements_for_metrics(
                metric_references=tuple(MetricReference(element_name=mname) for mname in metric_names),
                element_set_filter=LinkableElementFilter(
                    without_any_of=frozenset(without_any_property),
                ),
            )
        ).path_key_to_linkable_dimensions

        dimensions: List[Dimension] = []
        linkable_dimensions_tuple: Tuple[LinkableDimension, ...]
        for (
            path_key,
            linkable_dimensions_tuple,
        ) in path_key_to_linkable_dimensions.items():
            for linkable_dimension in linkable_dimensions_tuple:
                # Simple dimensions shouldn't show date part items.
                if linkable_dimension.date_part is not None:
                    continue

                if LinkableElementProperty.METRIC_TIME in linkable_dimension.properties:
                    metric_time_name = DataSet.metric_time_dimension_name()
                    assert linkable_dimension.element_name == metric_time_name, (
                        f"{linkable_dimension} has the {LinkableElementProperty.METRIC_TIME}, but the name does not"
                        f"match."
                    )

                    dimensions.append(
                        Dimension(
                            name=metric_time_name,
                            qualified_name=StructuredLinkableSpecName(
                                element_name=metric_time_name,
                                entity_link_names=tuple(
                                    entity_reference.element_name
                                    for entity_reference in linkable_dimension.entity_links
                                ),
                                time_granularity_name=(
                                    linkable_dimension.time_granularity.name
                                    if linkable_dimension.time_granularity is not None
                                    else None
                                ),
                            ).qualified_name,
                            entity_links=(),
                            description="Event time for metrics.",
                            metadata=None,
                            type_params=PydanticDimensionTypeParams(
                                time_granularity=(
                                    linkable_dimension.time_granularity.base_granularity
                                    if linkable_dimension.time_granularity is not None
                                    else None
                                ),
                                validity_params=None,
                            ),
                            is_partition=False,
                            type=DimensionType.TIME,
                        )
                    )
                else:
                    assert (
                        linkable_dimension.defined_in_semantic_model
                    ), "Only metric_time can have no semantic_model_origin."
                    semantic_model = self._semantic_manifest_lookup.semantic_model_lookup.get_by_reference(
                        linkable_dimension.defined_in_semantic_model
                    )
                    assert semantic_model
                    dimensions.append(
                        Dimension.from_pydantic(
                            pydantic_dimension=SemanticModelHelper.get_dimension_from_semantic_model(
                                semantic_model=semantic_model,
                                dimension_reference=linkable_dimension.reference,
                            ),
                            entity_links=path_key.entity_links,
                        )
                    )
        return sorted(dimensions, key=lambda dimension: dimension.qualified_name)

    def list_dimensions(self) -> List[Dimension]:  # noqa: D102
        """Get full dimension object for all dimensions in the semantic manifest."""
        semantic_model_lookup = self._semantic_manifest_lookup.semantic_model_lookup

        dimensions: List[Dimension] = []
        for dimension_reference in semantic_model_lookup.get_dimension_references():
            for semantic_model in semantic_model_lookup.get_semantic_models_for_dimension(dimension_reference):
                dimensions.append(
                    Dimension.from_pydantic(
                        pydantic_dimension=SemanticModelHelper.get_dimension_from_semantic_model(
                            semantic_model=semantic_model, dimension_reference=dimension_reference
                        ),
                        entity_links=(SemanticModelHelper.resolved_primary_entity(semantic_model),),
                    )
                )

        return dimensions

    def entities_for_metrics(self, metric_names: List[str]) -> List[Entity]:  # noqa: D102
        path_key_to_linkable_entities = (
            self._semantic_manifest_lookup.metric_lookup.linkable_elements_for_metrics(
                metric_references=tuple(MetricReference(element_name=mname) for mname in metric_names),
                element_set_filter=LinkableElementFilter(
                    with_any_of=frozenset(
                        {
                            LinkableElementProperty.ENTITY,
                        }
                    ),
                ),
            )
        ).path_key_to_linkable_entities

        entities: List[Entity] = []
        for (
            path_key,
            linkable_entity_tuple,
        ) in path_key_to_linkable_entities.items():
            for linkable_entity in linkable_entity_tuple:
                semantic_model = self._semantic_manifest_lookup.semantic_model_lookup.get_by_reference(
                    linkable_entity.defined_in_semantic_model
                )
                assert semantic_model
                entities.append(
                    Entity.from_pydantic(
                        pydantic_entity=SemanticModelHelper.get_entity_from_semantic_model(
                            semantic_model=semantic_model,
                            entity_reference=EntityReference(element_name=linkable_entity.element_name),
                        )
                    )
                )
        return entities

    @log_call(module_name=__name__, telemetry_reporter=_telemetry_reporter)
    def list_metrics(self) -> List[Metric]:  # noqa: D102
        metric_references = self._semantic_manifest_lookup.metric_lookup.metric_references
        metrics = self._semantic_manifest_lookup.metric_lookup.get_metrics(metric_references)
        return [
            Metric.from_pydantic(
                pydantic_metric=metric,
                dimensions=self.simple_dimensions_for_metrics([metric.name]),
            )
            for metric in metrics
        ]

    @log_call(module_name=__name__, telemetry_reporter=_telemetry_reporter)
    def list_saved_queries(self) -> List[SavedQuery]:  # noqa: D102
        return [
            SavedQuery.from_pydantic(saved_query)
            for saved_query in self._semantic_manifest_lookup.semantic_manifest.saved_queries
        ]

    @log_call(module_name=__name__, telemetry_reporter=_telemetry_reporter)
    def get_dimension_values(  # noqa: D102
        self,
        metric_names: List[str],
        get_group_by_values: str,
        time_constraint_start: Optional[datetime.datetime] = None,
        time_constraint_end: Optional[datetime.datetime] = None,
    ) -> List[str]:
        # Run query
        query_result: MetricFlowQueryResult = self.query(
            MetricFlowQueryRequest.create_with_random_request_id(
                metric_names=metric_names,
                group_by_names=[get_group_by_values],
                time_constraint_start=time_constraint_start,
                time_constraint_end=time_constraint_end,
                query_type=MetricFlowQueryType.DIMENSION_VALUES,
            )
        )
        if query_result.result_df is None:
            return []

        return sorted([str(val) for val in query_result.result_df.column_values_iterator(0)])

    @log_call(module_name=__name__, telemetry_reporter=_telemetry_reporter)
    def explain_get_dimension_values(  # noqa: D102
        self,
        metric_names: Optional[List[str]] = None,
        metrics: Optional[Sequence[MetricQueryParameter]] = None,
        get_group_by_values: Optional[str] = None,
        group_by: Optional[GroupByParameter] = None,
        time_constraint_start: Optional[datetime.datetime] = None,
        time_constraint_end: Optional[datetime.datetime] = None,
        min_max_only: bool = False,
    ) -> MetricFlowExplainResult:
        assert not (
            get_group_by_values and group_by
        ), "Both get_group_by_values and group_by were set, but if a group by is specified you should only use one of these!"
        return self._create_execution_plan(
            MetricFlowQueryRequest.create_with_random_request_id(
                metric_names=metric_names,
                metrics=metrics,
                group_by_names=(get_group_by_values,) if get_group_by_values else None,
                group_by=(group_by,) if group_by else None,
                time_constraint_start=time_constraint_start,
                time_constraint_end=time_constraint_end,
                min_max_only=min_max_only,
                query_type=MetricFlowQueryType.DIMENSION_VALUES,
            )
        )



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional, Sequence, Tuple

from metricflow_semantics.dag.id_prefix import IdPrefix, StaticIdPrefix
from metricflow_semantics.dag.mf_dag import DisplayedProperty
from metricflow_semantics.mf_logging.pretty_print import mf_pformat
from metricflow_semantics.specs.dunder_column_association_resolver import DunderColumnAssociationResolver
from metricflow_semantics.specs.spec_set import InstanceSpecSet
from metricflow_semantics.visitor import VisitorOutputT

from metricflow.dataflow.dataflow_plan import DataflowPlanNode
from metricflow.dataflow.dataflow_plan_visitor import DataflowPlanNodeVisitor


@dataclass(frozen=True, eq=False)
class FilterElementsNode(DataflowPlanNode):
    """Only passes the listed elements.

    Attributes:
        include_specs: The specs for the elements that it should pass.
        replace_description: Replace the default description with this.
        distinct: If you only want the distinct values for the selected specs..
    """

    include_specs: InstanceSpecSet
    replace_description: Optional[str] = None
    distinct: bool = False

    def __post_init__(self) -> None:  # noqa: D105
        super().__post_init__()
        assert len(self.parent_nodes) == 1

    @staticmethod
    def create(  # noqa: D102
        parent_node: DataflowPlanNode,
        include_specs: InstanceSpecSet,
        replace_description: Optional[str] = None,
        distinct: bool = False,
    ) -> FilterElementsNode:
        return FilterElementsNode(
            parent_nodes=(parent_node,),
            include_specs=include_specs,
            replace_description=replace_description,
            distinct=distinct,
        )

    @classmethod
    def id_prefix(cls) -> IdPrefix:  # noqa: D102
        return StaticIdPrefix.DATAFLOW_NODE_PASS_FILTER_ELEMENTS_ID_PREFIX

    def accept(self, visitor: DataflowPlanNodeVisitor[VisitorOutputT]) -> VisitorOutputT:  # noqa: D102
        return visitor.visit_filter_elements_node(self)

    @property
    def description(self) -> str:  # noqa: D102
        if self.replace_description:
            return self.replace_description

        column_resolver = DunderColumnAssociationResolver()
        return f"Pass Only Elements: {mf_pformat([column_resolver.resolve_spec(spec).column_name for spec in self.include_specs.all_specs])}"

    @property
    def displayed_properties(self) -> Sequence[DisplayedProperty]:  # noqa: D102
        additional_properties: Tuple[DisplayedProperty, ...] = ()
        if not self.replace_description:
            additional_properties = tuple(
                DisplayedProperty("include_spec", include_spec) for include_spec in self.include_specs.all_specs
            ) + (
                DisplayedProperty("distinct", self.distinct),
            )
        return tuple(super().displayed_properties) + additional_properties

    @property
    def parent_node(self) -> DataflowPlanNode:  # noqa: D102
        return self.parent_nodes[0]

    def functionally_identical(self, other_node: DataflowPlanNode) -> bool:  # noqa: D102
        return (
            isinstance(other_node, self.__class__)
            and other_node.include_specs == self.include_specs
            and other_node.distinct == self.distinct
        )

    def with_new_parents(self, new_parent_nodes: Sequence[DataflowPlanNode]) -> FilterElementsNode:  # noqa: D102
        assert len(new_parent_nodes) == 1
        return FilterElementsNode(
            parent_nodes=tuple(new_parent_nodes),
            include_specs=self.include_specs,
            distinct=self.distinct,
            replace_description=self.replace_description,
        )



from __future__ import annotations

import logging
from typing import Sequence

from dbt_semantic_interfaces.protocols import WhereFilterIntersection
from dbt_semantic_interfaces.references import MeasureReference, MetricReference
from typing_extensions import override

from metricflow_semantics.query.group_by_item.resolution_path import MetricFlowQueryResolutionPath
from metricflow_semantics.query.issues.issues_base import MetricFlowQueryResolutionIssueSet
from metricflow_semantics.query.issues.parsing.duplicate_metric import DuplicateMetricIssue
from metricflow_semantics.query.validation_rules.base_validation_rule import PostResolutionQueryValidationRule

logger = logging.getLogger(__name__)


class DuplicateMetricValidationRule(PostResolutionQueryValidationRule):
    """Validates that a query does not include the same metric multiple times."""

    @override
    def validate_metric_in_resolution_dag(
        self,
        metric_reference: MetricReference,
        resolution_path: MetricFlowQueryResolutionPath,
    ) -> MetricFlowQueryResolutionIssueSet:
        return MetricFlowQueryResolutionIssueSet.empty_instance()

    @override
    def validate_query_in_resolution_dag(
        self,
        metrics_in_query: Sequence[MetricReference],
        where_filter_intersection: WhereFilterIntersection,
        resolution_path: MetricFlowQueryResolutionPath,
    ) -> MetricFlowQueryResolutionIssueSet:
        duplicate_metric_references = tuple(
            sorted(
                metric_reference
                for metric_reference in set(metrics_in_query)
                if metrics_in_query.count(metric_reference) > 1
            )
        )

        if len(duplicate_metric_references) > 0:
            return MetricFlowQueryResolutionIssueSet.from_issue(
                DuplicateMetricIssue.from_parameters(
                    query_resolution_path=resolution_path, duplicate_metric_references=duplicate_metric_references
                )
            )

        return MetricFlowQueryResolutionIssueSet.empty_instance()

    @override
    def validate_measure_in_resolution_dag(
        self,
        measure_reference: MeasureReference,
        resolution_path: MetricFlowQueryResolutionPath,
    ) -> MetricFlowQueryResolutionIssueSet:
        return MetricFlowQueryResolutionIssueSet.empty_instance()



from __future__ import annotations

from enum import Enum
from typing import Any, FrozenSet


class LinkableElementProperty(Enum):
    """The properties associated with a valid linkable element.

    Local means an element that is defined within the same semantic model as the measure. This definition is used
    throughout the related classes.
    """

    # A local element as per above definition.
    LOCAL = "local"
    # A local dimension that is prefixed with a local primary entity.
    LOCAL_LINKED = "local_linked"
    # An element that was joined to the measure semantic model by an entity.
    JOINED = "joined"
    # An element that was joined to the measure semantic model by joining multiple semantic models.
    MULTI_HOP = "multi_hop"
    # A time dimension that is a version of a time dimension in a semantic model, but at a different granularity.
    DERIVED_TIME_GRANULARITY = "derived_time_granularity"
    # Refers to an entity, not a dimension.
    ENTITY = "entity"
    # See metric_time in DataSet
    METRIC_TIME = "metric_time"
    # Refers to a metric, not a dimension.
    METRIC = "metric"
    # A time dimension with a DatePart.
    DATE_PART = "date_part"
    # A linkable element that is itself part of an SCD model, or a linkable element that gets joined through another SCD model.
    SCD_HOP = "scd_hop"

    @staticmethod
    def all_properties() -> FrozenSet[LinkableElementProperty]:  # noqa: D102
        return frozenset({linkable_element_property for linkable_element_property in LinkableElementProperty})

    def __lt__(self, other: Any) -> bool:  # type: ignore[misc]
        """When ordering, order by the enum name."""
        if not isinstance(other, LinkableElementProperty):
            return NotImplemented

        return self.name < other.name



from __future__ import annotations

import functools
from dataclasses import dataclass
from enum import Enum
from typing import Tuple

from dbt_semantic_interfaces.enum_extension import assert_values_exhausted

from metricflow.sql.optimizer.column_pruner import SqlColumnPrunerOptimizer
from metricflow.sql.optimizer.rewriting_sub_query_reducer import SqlRewritingSubQueryReducer
from metricflow.sql.optimizer.sql_query_plan_optimizer import SqlPlanOptimizer
from metricflow.sql.optimizer.sub_query_reducer import SqlSubQueryReducer
from metricflow.sql.optimizer.table_alias_simplifier import SqlTableAliasSimplifier


@functools.total_ordering
class SqlOptimizationLevel(Enum):
    """Defines the level of query optimization and the associated optimizers to apply."""

    O0 = "O0"
    O1 = "O1"
    O2 = "O2"
    O3 = "O3"
    O4 = "O4"
    O5 = "O5"

    @staticmethod
    def default_level() -> SqlOptimizationLevel:  # noqa: D102
        return SqlOptimizationLevel.O5

    def __lt__(self, other: SqlOptimizationLevel) -> bool:  # noqa: D105
        if not isinstance(other, SqlOptimizationLevel):
            return NotImplemented

        return self.name < other.name


@dataclass(frozen=True)
class SqlGenerationOptionSet:
    """Defines the different SQL generation optimizers / options that should be used at each level."""

    optimizers: Tuple[SqlPlanOptimizer, ...]

    # Specifies whether CTEs can be used to simplify generated SQL.
    allow_cte: bool

    @staticmethod
    def options_for_level(  # noqa: D102
        level: SqlOptimizationLevel, use_column_alias_in_group_by: bool
    ) -> SqlGenerationOptionSet:
        optimizers: Tuple[SqlPlanOptimizer, ...] = ()
        allow_cte = False
        if level is SqlOptimizationLevel.O0:
            pass
        elif level is SqlOptimizationLevel.O1:
            optimizers = (SqlTableAliasSimplifier(),)
        elif level is SqlOptimizationLevel.O2:
            optimizers = (SqlColumnPrunerOptimizer(), SqlTableAliasSimplifier())
        elif level is SqlOptimizationLevel.O3:
            optimizers = (SqlColumnPrunerOptimizer(), SqlSubQueryReducer(), SqlTableAliasSimplifier())
        elif level is SqlOptimizationLevel.O4:
            optimizers = (
                SqlColumnPrunerOptimizer(),
                SqlRewritingSubQueryReducer(use_column_alias_in_group_bys=use_column_alias_in_group_by),
                SqlTableAliasSimplifier(),
            )
        elif level is SqlOptimizationLevel.O5:
            optimizers = (
                SqlColumnPrunerOptimizer(),
                SqlRewritingSubQueryReducer(use_column_alias_in_group_bys=use_column_alias_in_group_by),
                SqlTableAliasSimplifier(),
            )
            allow_cte = True
        else:
            assert_values_exhausted(level)

        return SqlGenerationOptionSet(
            optimizers=optimizers,
            allow_cte=allow_cte,
        )



from __future__ import annotations

from dataclasses import dataclass
from typing import Tuple

from dbt_semantic_interfaces.dataclass_serialization import SerializableDataclass
from dbt_semantic_interfaces.references import DimensionReference, EntityReference
from typing_extensions import override

from metricflow_semantics.model.semantics.linkable_element import ElementPathKey, LinkableElementType
from metricflow_semantics.specs.instance_spec import InstanceSpecVisitor, LinkableInstanceSpec
from metricflow_semantics.visitor import VisitorOutputT


@dataclass(frozen=True)
class DimensionSpec(LinkableInstanceSpec, SerializableDataclass):  # noqa: D101
    element_name: str
    entity_links: Tuple[EntityReference, ...]

    @property
    def without_first_entity_link(self) -> DimensionSpec:  # noqa: D102
        assert len(self.entity_links) > 0, f"Spec does not have any entity links: {self}"
        return DimensionSpec(element_name=self.element_name, entity_links=self.entity_links[1:])

    @property
    def without_entity_links(self) -> DimensionSpec:  # noqa: D102
        return DimensionSpec(element_name=self.element_name, entity_links=())

    @staticmethod
    def from_linkable(spec: LinkableInstanceSpec) -> DimensionSpec:  # noqa: D102
        return DimensionSpec(element_name=spec.element_name, entity_links=spec.entity_links)

    @property
    def reference(self) -> DimensionReference:  # noqa: D102
        return DimensionReference(element_name=self.element_name)

    def accept(self, visitor: InstanceSpecVisitor[VisitorOutputT]) -> VisitorOutputT:  # noqa: D102
        return visitor.visit_dimension_spec(self)

    @property
    @override
    def element_path_key(self) -> ElementPathKey:
        return ElementPathKey(
            element_name=self.element_name, element_type=LinkableElementType.DIMENSION, entity_links=self.entity_links
        )

    def with_entity_prefix(self, entity_prefix: EntityReference) -> DimensionSpec:  # noqa: D102
        return DimensionSpec(element_name=self.element_name, entity_links=(entity_prefix,) + self.entity_links)



from __future__ import annotations

from typing import Collection

from dateutil.parser import parse
from dbt_semantic_interfaces.enum_extension import assert_values_exhausted
from dbt_semantic_interfaces.type_enums.date_part import DatePart
from dbt_semantic_interfaces.type_enums.time_granularity import TimeGranularity
from metricflow_semantics.sql.sql_bind_parameters import SqlBindParameterSet
from metricflow_semantics.sql.sql_exprs import (
    SqlAddTimeExpression,
    SqlArithmeticExpression,
    SqlArithmeticOperator,
    SqlBetweenExpression,
    SqlGenerateUuidExpression,
    SqlIntegerExpression,
    SqlPercentileExpression,
    SqlPercentileFunctionType,
    SqlSubtractTimeIntervalExpression,
)
from typing_extensions import override

from metricflow.protocols.sql_client import SqlEngine
from metricflow.sql.render.expr_renderer import (
    DefaultSqlExpressionRenderer,
    SqlExpressionRenderer,
    SqlExpressionRenderResult,
)
from metricflow.sql.render.sql_plan_renderer import DefaultSqlPlanRenderer


class TrinoSqlExpressionRenderer(DefaultSqlExpressionRenderer):
    """Expression renderer for the Trino engine."""

    sql_engine = SqlEngine.TRINO

    @property
    @override
    def supported_percentile_function_types(self) -> Collection[SqlPercentileFunctionType]:
        return {
            SqlPercentileFunctionType.APPROXIMATE_CONTINUOUS,
        }

    @override
    def visit_generate_uuid_expr(self, node: SqlGenerateUuidExpression) -> SqlExpressionRenderResult:
        return SqlExpressionRenderResult(
            sql="uuid()",
            bind_parameter_set=SqlBindParameterSet(),
        )

    @override
    def visit_subtract_time_interval_expr(self, node: SqlSubtractTimeIntervalExpression) -> SqlExpressionRenderResult:
        """Render time delta for Trino, require granularity in quotes and function name change."""
        arg_rendered = node.arg.accept(self)

        count = node.count
        granularity = node.granularity
        if granularity is TimeGranularity.QUARTER:
            granularity = TimeGranularity.MONTH
            count *= 3
        return SqlExpressionRenderResult(
            sql=f"DATE_ADD('{granularity.value}', -{count}, {arg_rendered.sql})",
            bind_parameter_set=arg_rendered.bind_parameter_set,
        )

    @override
    def visit_add_time_expr(self, node: SqlAddTimeExpression) -> SqlExpressionRenderResult:
        """Render time delta for Trino, require granularity in quotes and function name change."""
        granularity = node.granularity
        count_expr = node.count_expr
        if granularity is TimeGranularity.QUARTER:
            granularity = TimeGranularity.MONTH
            SqlArithmeticExpression.create(
                left_expr=node.count_expr,
                operator=SqlArithmeticOperator.MULTIPLY,
                right_expr=SqlIntegerExpression.create(3),
            )

        arg_rendered = node.arg.accept(self)
        count_rendered = count_expr.accept(self)
        count_sql = f"({count_rendered.sql})" if count_expr.requires_parenthesis else count_rendered.sql

        return SqlExpressionRenderResult(
            sql=f"DATE_ADD('{granularity.value}', {count_sql}, {arg_rendered.sql})",
            bind_parameter_set=SqlBindParameterSet.merge_iterable(
                (arg_rendered.bind_parameter_set, count_rendered.bind_parameter_set)
            ),
        )

    @override
    def visit_percentile_expr(self, node: SqlPercentileExpression) -> SqlExpressionRenderResult:
        """Render a percentile expression for Trino."""
        arg_rendered = self.render_sql_expr(node.order_by_arg)
        params = arg_rendered.bind_parameter_set
        percentile = node.percentile_args.percentile

        if node.percentile_args.function_type is SqlPercentileFunctionType.APPROXIMATE_CONTINUOUS:
            return SqlExpressionRenderResult(
                sql=f"approx_percentile({arg_rendered.sql}, {percentile})",
                bind_parameter_set=params,
            )
        elif (
            node.percentile_args.function_type is SqlPercentileFunctionType.APPROXIMATE_DISCRETE
            or node.percentile_args.function_type is SqlPercentileFunctionType.DISCRETE
            or node.percentile_args.function_type is SqlPercentileFunctionType.CONTINUOUS
        ):
            raise RuntimeError(
                "Discrete, Continuous and Approximate discrete percentile aggregates are not supported for Trino. Set "
                + "use_approximate_percentile and disable use_discrete_percentile in all percentile measures."
            )
        else:
            assert_values_exhausted(node.percentile_args.function_type)

    @override
    def visit_between_expr(self, node: SqlBetweenExpression) -> SqlExpressionRenderResult:
        """Render a between expression for Trino. If the expression is a timestamp literal then wrap literals with timestamp."""
        rendered_column_arg = self.render_sql_expr(node.column_arg)
        rendered_start_expr = self.render_sql_expr(node.start_expr)
        rendered_end_expr = self.render_sql_expr(node.end_expr)

        bind_parameter_set = SqlBindParameterSet()
        bind_parameter_set = bind_parameter_set.merge(rendered_column_arg.bind_parameter_set)
        bind_parameter_set = bind_parameter_set.merge(rendered_start_expr.bind_parameter_set)
        bind_parameter_set = bind_parameter_set.merge(rendered_end_expr.bind_parameter_set)

        # Handle timestamp literals differently.
        if parse(rendered_start_expr.sql):
            sql = f"{rendered_column_arg.sql} BETWEEN timestamp {rendered_start_expr.sql} AND timestamp {rendered_end_expr.sql}"
        else:
            sql = f"{rendered_column_arg.sql} BETWEEN {rendered_start_expr.sql} AND {rendered_end_expr.sql}"

        return SqlExpressionRenderResult(
            sql=sql,
            bind_parameter_set=bind_parameter_set,
        )

    @override
    def render_date_part(self, date_part: DatePart) -> str:
        """Render DATE PART for an EXTRACT expression.

        Override DAY_OF_WEEK in Trino to ISO date part to ensure all engines return consistent results.
        """
        if date_part is DatePart.DOW:
            return "DAY_OF_WEEK"

        return date_part.value


class TrinoSqlPlanRenderer(DefaultSqlPlanRenderer):
    """Plan renderer for the Trino engine."""

    EXPR_RENDERER = TrinoSqlExpressionRenderer()

    @property
    @override
    def expr_renderer(self) -> SqlExpressionRenderer:
        return self.EXPR_RENDERER



from __future__ import annotations

from typing import Collection

from dbt_semantic_interfaces.enum_extension import assert_values_exhausted
from dbt_semantic_interfaces.type_enums.time_granularity import TimeGranularity
from metricflow_semantics.errors.error_classes import UnsupportedEngineFeatureError
from metricflow_semantics.sql.sql_bind_parameters import SqlBindParameterSet
from metricflow_semantics.sql.sql_exprs import (
    SqlAddTimeExpression,
    SqlArithmeticExpression,
    SqlArithmeticOperator,
    SqlGenerateUuidExpression,
    SqlIntegerExpression,
    SqlPercentileExpression,
    SqlPercentileFunctionType,
    SqlSubtractTimeIntervalExpression,
)
from typing_extensions import override

from metricflow.protocols.sql_client import SqlEngine
from metricflow.sql.render.expr_renderer import (
    DefaultSqlExpressionRenderer,
    SqlExpressionRenderer,
    SqlExpressionRenderResult,
)
from metricflow.sql.render.sql_plan_renderer import DefaultSqlPlanRenderer


class PostgresSqlExpressionRenderer(DefaultSqlExpressionRenderer):
    """Expression renderer for the PostgreSQL engine."""

    sql_engine = SqlEngine.POSTGRES

    @property
    @override
    def double_data_type(self) -> str:
        """Custom double data type for the PostgreSQL engine."""
        return "DOUBLE PRECISION"

    @property
    @override
    def supported_percentile_function_types(self) -> Collection[SqlPercentileFunctionType]:
        return {SqlPercentileFunctionType.CONTINUOUS, SqlPercentileFunctionType.DISCRETE}

    @override
    def visit_subtract_time_interval_expr(self, node: SqlSubtractTimeIntervalExpression) -> SqlExpressionRenderResult:
        """Render time delta operations for PostgreSQL, which needs custom support for quarterly granularity."""
        arg_rendered = node.arg.accept(self)

        count = node.count
        granularity = node.granularity
        if granularity is TimeGranularity.QUARTER:
            granularity = TimeGranularity.MONTH
            count *= 3
        return SqlExpressionRenderResult(
            sql=f"{arg_rendered.sql} - MAKE_INTERVAL({granularity.value}s => {count})",
            bind_parameter_set=arg_rendered.bind_parameter_set,
        )

    @override
    def visit_add_time_expr(self, node: SqlAddTimeExpression) -> SqlExpressionRenderResult:
        """Render time delta operations for PostgreSQL, which needs custom support for quarterly granularity."""
        granularity = node.granularity
        count_expr = node.count_expr
        if granularity is TimeGranularity.QUARTER:
            granularity = TimeGranularity.MONTH
            SqlArithmeticExpression.create(
                left_expr=node.count_expr,
                operator=SqlArithmeticOperator.MULTIPLY,
                right_expr=SqlIntegerExpression.create(3),
            )

        arg_rendered = node.arg.accept(self)
        count_rendered = count_expr.accept(self)
        count_sql = f"({count_rendered.sql})" if count_expr.requires_parenthesis else count_rendered.sql

        return SqlExpressionRenderResult(
            sql=f"{arg_rendered.sql} + MAKE_INTERVAL({granularity.value}s => CAST ({count_sql} AS INTEGER))",
            bind_parameter_set=SqlBindParameterSet.merge_iterable(
                (arg_rendered.bind_parameter_set, count_rendered.bind_parameter_set)
            ),
        )

    @override
    def visit_generate_uuid_expr(self, node: SqlGenerateUuidExpression) -> SqlExpressionRenderResult:
        return SqlExpressionRenderResult(
            sql="GEN_RANDOM_UUID()",
            bind_parameter_set=SqlBindParameterSet(),
        )

    @override
    def visit_percentile_expr(self, node: SqlPercentileExpression) -> SqlExpressionRenderResult:
        """Render a percentile expression for Postgres."""
        arg_rendered = self.render_sql_expr(node.order_by_arg)
        params = arg_rendered.bind_parameter_set
        percentile = node.percentile_args.percentile

        if node.percentile_args.function_type is SqlPercentileFunctionType.CONTINUOUS:
            function_str = "PERCENTILE_CONT"
        elif node.percentile_args.function_type is SqlPercentileFunctionType.DISCRETE:
            function_str = "PERCENTILE_DISC"
        elif node.percentile_args.function_type is SqlPercentileFunctionType.APPROXIMATE_CONTINUOUS:
            raise UnsupportedEngineFeatureError(
                "Approximate continuous percentile aggregate not supported for Postgres. Set "
                + "use_approximate_percentile to false in all percentile measures."
            )
        elif node.percentile_args.function_type is SqlPercentileFunctionType.APPROXIMATE_DISCRETE:
            raise UnsupportedEngineFeatureError(
                "Approximate discrete percentile aggregate not supported for Postgres. Set "
                + "use_approximate_percentile to false in all percentile measures."
            )
        else:
            assert_values_exhausted(node.percentile_args.function_type)

        return SqlExpressionRenderResult(
            sql=f"{function_str}({percentile}) WITHIN GROUP (ORDER BY ({arg_rendered.sql}))",
            bind_parameter_set=params,
        )


class PostgresSQLSqlPlanRenderer(DefaultSqlPlanRenderer):
    """Plan renderer for the PostgreSQL engine."""

    EXPR_RENDERER = PostgresSqlExpressionRenderer()

    @property
    @override
    def expr_renderer(self) -> SqlExpressionRenderer:
        return self.EXPR_RENDERER



from __future__ import annotations

import traceback
from dataclasses import dataclass

from dbt_semantic_interfaces.protocols import WhereFilter
from typing_extensions import override

from metricflow_semantics.mf_logging.formatting import indent
from metricflow_semantics.query.group_by_item.resolution_path import MetricFlowQueryResolutionPath
from metricflow_semantics.query.issues.issues_base import (
    MetricFlowQueryIssueType,
    MetricFlowQueryResolutionIssue,
)
from metricflow_semantics.query.resolver_inputs.base_resolver_inputs import MetricFlowQueryResolverInput


@dataclass(frozen=True)
class WhereFilterParsingIssue(MetricFlowQueryResolutionIssue):
    """Describes an issue with a query where one of the where filters can't be parsed."""

    where_filter: WhereFilter
    parse_exception: Exception

    @staticmethod
    def from_parameters(  # noqa: D102
        where_filter: WhereFilter,
        parse_exception: Exception,
        query_resolution_path: MetricFlowQueryResolutionPath,
    ) -> WhereFilterParsingIssue:
        return WhereFilterParsingIssue(
            issue_type=MetricFlowQueryIssueType.ERROR,
            parent_issues=(),
            query_resolution_path=query_resolution_path,
            where_filter=where_filter,
            parse_exception=parse_exception,
        )

    @override
    def ui_description(self, associated_input: MetricFlowQueryResolverInput) -> str:
        return (
            f"Error parsing where filter:\n\n"
            f"{indent(repr(self.where_filter.where_sql_template))}\n\n"
            f"Got exception:\n\n"
            f"{indent(''.join(traceback.TracebackException.from_exception(self.parse_exception).format()))}"
        )

    @override
    def with_path_prefix(self, path_prefix: MetricFlowQueryResolutionPath) -> WhereFilterParsingIssue:
        return WhereFilterParsingIssue(
            issue_type=self.issue_type,
            parent_issues=tuple(issue.with_path_prefix(path_prefix) for issue in self.parent_issues),
            query_resolution_path=self.query_resolution_path.with_path_prefix(path_prefix),
            where_filter=self.where_filter,
            parse_exception=self.parse_exception,
        )



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional, Sequence, Sized, Tuple

from dbt_semantic_interfaces.references import SemanticModelReference
from typing_extensions import override

from metricflow_semantics.collection_helpers.merger import Mergeable
from metricflow_semantics.mf_logging.pretty_print import mf_pformat
from metricflow_semantics.model.semantic_model_derivation import SemanticModelDerivation
from metricflow_semantics.query.group_by_item.filter_spec_resolution.filter_spec_lookup import (
    FilterSpecResolutionLookUp,
)
from metricflow_semantics.query.group_by_item.resolution_dag.dag import GroupByItemResolutionDag
from metricflow_semantics.query.issues.issues_base import MetricFlowQueryResolutionIssueSet
from metricflow_semantics.query.resolver_inputs.base_resolver_inputs import MetricFlowQueryResolverInput
from metricflow_semantics.specs.query_spec import MetricFlowQuerySpec


@dataclass(frozen=True)
class InputToIssueSetMappingItem:
    """A mapping item that associates an input to issues generated when resolving the input."""

    resolver_input: MetricFlowQueryResolverInput
    issue_set: MetricFlowQueryResolutionIssueSet


@dataclass(frozen=True)
class InputToIssueSetMapping(Mergeable, Sized):
    """Collection of InputToIssueSetMappingItems.

    This implements Sized to aid mf_pformat(). Sized is used in mf_pformat() when excluding display of empty items.
    """

    items: Tuple[InputToIssueSetMappingItem, ...]

    @property
    def has_issues(self) -> bool:  # noqa: D102
        return any(item.issue_set.has_issues for item in self.items)

    @property
    def merged_issue_set(self) -> MetricFlowQueryResolutionIssueSet:
        """Return an issue set that contains all issues in the included mapping items."""
        return MetricFlowQueryResolutionIssueSet.merge_iterable(item.issue_set for item in self.items)

    @override
    def merge(self, other: InputToIssueSetMapping) -> InputToIssueSetMapping:
        return InputToIssueSetMapping(items=self.items + other.items)

    @classmethod
    @override
    def empty_instance(cls) -> InputToIssueSetMapping:
        return InputToIssueSetMapping(
            items=(),
        )

    @staticmethod
    def from_one_item(  # noqa: D102
        resolver_input: MetricFlowQueryResolverInput, issue_set: MetricFlowQueryResolutionIssueSet
    ) -> InputToIssueSetMapping:
        return InputToIssueSetMapping(
            items=(InputToIssueSetMappingItem(resolver_input=resolver_input, issue_set=issue_set),)
        )

    @override
    def __len__(self) -> int:
        return len(self.items)


@dataclass(frozen=True)
class MetricFlowQueryResolution(SemanticModelDerivation):
    """The result of resolving query inputs to specs."""

    # Can be None if there were errors.
    query_spec: Optional[MetricFlowQuerySpec]
    # The resolution DAG generated for the query.
    resolution_dag: Optional[GroupByItemResolutionDag]
    # The lookup that is used later in the DataflowPlanBuilder to figure out which specs are required by the filters
    # in the queries / metrics.
    filter_spec_lookup: FilterSpecResolutionLookUp
    # Mapping of issues with the inputs.
    input_to_issue_set: InputToIssueSetMapping
    # The semantic models that would be queried to resolve the query.
    queried_semantic_models: Tuple[SemanticModelReference, ...]

    @property
    @override
    def derived_from_semantic_models(self) -> Sequence[SemanticModelReference]:
        return self.queried_semantic_models

    @property
    def checked_query_spec(self) -> MetricFlowQuerySpec:
        """Returns the query_spec, but if MetricFlowQueryResolution.has_errors was True, raise a RuntimeError."""
        if self.input_to_issue_set.has_issues:
            raise RuntimeError(
                f"Can't get the query spec because errors were present in the resolution:\n"
                f"{mf_pformat(self.input_to_issue_set.has_issues)}"
            )
        if self.query_spec is None:
            raise RuntimeError("If there were no errors, query_spec should have been populated.")
        return self.query_spec

    @property
    def has_errors(self) -> bool:  # noqa: D102
        return self.input_to_issue_set.has_issues or self.filter_spec_lookup.has_errors



from __future__ import annotations

from dataclasses import dataclass
from typing import List, Tuple

from metricflow_semantics.specs.linkable_spec_set import LinkableSpecSet

from metricflow.dataflow.builder.node_evaluator import JoinLinkableInstancesRecipe
from metricflow.dataflow.dataflow_plan import DataflowPlanNode
from metricflow.dataflow.nodes.join_to_base import JoinDescription


@dataclass(frozen=True)
class SourceNodeRecipe:
    """Get a recipe for how to build a dataflow plan node that outputs measures and linkable instances as needed."""

    source_node: DataflowPlanNode
    required_local_linkable_specs: LinkableSpecSet
    join_linkable_instances_recipes: Tuple[JoinLinkableInstancesRecipe, ...]
    all_linkable_specs_required_for_source_nodes: LinkableSpecSet

    @property
    def join_targets(self) -> List[JoinDescription]:
        """Joins to be made to source node."""
        return [join_recipe.join_description for join_recipe in self.join_linkable_instances_recipes]



from __future__ import annotations

import logging
from dataclasses import dataclass
from enum import Enum
from typing import Any, FrozenSet, List, Optional, Sequence, Tuple

from dbt_semantic_interfaces.references import EntityReference
from dbt_semantic_interfaces.type_enums.date_part import DatePart
from more_itertools import is_sorted
from typing_extensions import override

from metricflow_semantics.model.linkable_element_property import LinkableElementProperty
from metricflow_semantics.model.semantics.element_filter import LinkableElementFilter
from metricflow_semantics.specs.instance_spec import InstanceSpec, LinkableInstanceSpec
from metricflow_semantics.specs.patterns.spec_pattern import SpecPattern
from metricflow_semantics.specs.spec_set import group_specs_by_type

logger = logging.getLogger(__name__)


class ParameterSetField(Enum):
    """The fields of the SpecPatternParameterSet class used for matching in the EntityLinkPattern.

    Considering moving this to be a part of the specs module / classes.
    """

    ELEMENT_NAME = "element_name"
    ENTITY_LINKS = "entity_links"
    TIME_GRANULARITY = "time_granularity_name"
    DATE_PART = "date_part"
    METRIC_SUBQUERY_ENTITY_LINKS = "metric_subquery_entity_links"

    def __lt__(self, other: Any) -> bool:  # type: ignore[misc]
        """Allow for ordering so that a sequence of these can be consistently represented for test snapshots."""
        if self.__class__ is other.__class__:
            return self.value < other.value
        return NotImplemented


@dataclass(frozen=True)
class SpecPatternParameterSet:
    """See EntityPathPattern for more details."""

    # Specify the field values to compare. None can't be used to signal "don't compare" because sometimes a pattern
    # needs to match a spec where the field is None. These should be sorted.
    fields_to_compare: Tuple[ParameterSetField, ...]

    # The name of the element in the semantic model
    element_name: Optional[str] = None
    # The entities used for joining semantic models.
    entity_links: Optional[Tuple[EntityReference, ...]] = None
    # Properties of time dimensions to match.
    time_granularity_name: Optional[str] = None
    date_part: Optional[DatePart] = None
    metric_subquery_entity_links: Optional[Tuple[EntityReference, ...]] = None

    @staticmethod
    def from_parameters(  # noqa: D102
        fields_to_compare: Sequence[ParameterSetField],
        element_name: Optional[str] = None,
        entity_links: Optional[Sequence[EntityReference]] = None,
        time_granularity_name: Optional[str] = None,
        date_part: Optional[DatePart] = None,
        metric_subquery_entity_links: Optional[Tuple[EntityReference, ...]] = None,
    ) -> SpecPatternParameterSet:
        return SpecPatternParameterSet(
            fields_to_compare=tuple(sorted(fields_to_compare)),
            element_name=element_name,
            entity_links=tuple(entity_links) if entity_links is not None else None,
            time_granularity_name=time_granularity_name,
            date_part=date_part,
            metric_subquery_entity_links=metric_subquery_entity_links,
        )

    def __post_init__(self) -> None:
        """Check that fields_to_compare is sorted so that patterns that do the same thing can be compared."""
        assert is_sorted(self.fields_to_compare), "`fields_to_compare` must be sorted."


@dataclass(frozen=True)
class EntityLinkPattern(SpecPattern):
    """A pattern that matches group-by-items using the entity-link-path specification.

    The entity link path specifies how a group-by-item for a metric query should be constructed. The group-by-item
    is obtained by joining the semantic model containing the measure to a semantic model containing the group-by-
    item using a specified entity. Additional semantic models can be joined using additional entities to obtain the
    group-by-item. The series of entities that are used form the entity path. Since the entity path does not specify
    which semantic models need to be used, additional resolution is done in later stages to generate the necessary SQL.

    The entity links that are specified is used as a suffix match.
    """

    parameter_set: SpecPatternParameterSet

    def _match_entity_links(self, candidate_specs: Sequence[LinkableInstanceSpec]) -> Sequence[LinkableInstanceSpec]:
        assert self.parameter_set.entity_links is not None
        num_links_to_check = len(self.parameter_set.entity_links)
        matching_specs: Sequence[LinkableInstanceSpec] = tuple(
            candidate_spec
            for candidate_spec in candidate_specs
            if (
                self.parameter_set.entity_links[-num_links_to_check:]
                == candidate_spec.entity_links[-num_links_to_check:]
            )
        )

        if len(matching_specs) <= 1:
            return matching_specs

        # If multiple match, then return only the ones with the shortest entity link path. There could be multiple
        # e.g. booking__listing__country and listing__country will match with listing__country.
        shortest_entity_link_length = min(len(matching_spec.entity_links) for matching_spec in matching_specs)
        return tuple(spec for spec in matching_specs if len(spec.entity_links) == shortest_entity_link_length)

    def _match_time_granularities(
        self, candidate_specs: Sequence[LinkableInstanceSpec]
    ) -> Sequence[LinkableInstanceSpec]:
        """Do a partial match on time granularities."""
        matching_specs: Sequence[LinkableInstanceSpec] = tuple(
            candidate_spec
            for candidate_spec in group_specs_by_type(candidate_specs).time_dimension_specs
            if candidate_spec.time_granularity.name
            == (self.parameter_set.time_granularity_name.lower() if self.parameter_set.time_granularity_name else None)
        )

        return matching_specs

    @override
    def match(self, candidate_specs: Sequence[InstanceSpec]) -> Sequence[LinkableInstanceSpec]:
        filtered_candidate_specs = group_specs_by_type(candidate_specs).linkable_specs
        # Checks that SpecPatternParameterSetField is valid wrt to the parameter set.

        # Entity links could be a partial match, so it's handled separately.
        if ParameterSetField.ENTITY_LINKS in self.parameter_set.fields_to_compare:
            filtered_candidate_specs = self._match_entity_links(filtered_candidate_specs)
        # Time granularity is handled separately because it's comparing a string input to an ExpandedTimeGranularity.
        if ParameterSetField.TIME_GRANULARITY in self.parameter_set.fields_to_compare:
            filtered_candidate_specs = self._match_time_granularities(filtered_candidate_specs)

        other_keys_to_check = set(
            field_to_compare.value for field_to_compare in self.parameter_set.fields_to_compare
        ).difference({ParameterSetField.ENTITY_LINKS.value, ParameterSetField.TIME_GRANULARITY.value})

        matching_specs: List[LinkableInstanceSpec] = []
        parameter_set_values = tuple(getattr(self.parameter_set, key_to_check) for key_to_check in other_keys_to_check)
        for spec in filtered_candidate_specs:
            spec_values = tuple(getattr(spec, key_to_check, None) for key_to_check in other_keys_to_check)
            if spec_values == parameter_set_values:
                matching_specs.append(spec)

        return matching_specs

    @property
    @override
    def element_pre_filter(self) -> LinkableElementFilter:
        element_names: Optional[FrozenSet[str]] = None
        if ParameterSetField.ELEMENT_NAME in self.parameter_set.fields_to_compare:
            element_names = frozenset({self.parameter_set.element_name}) if self.parameter_set.element_name else None
        if (
            self.parameter_set.metric_subquery_entity_links is None
            or len(self.parameter_set.metric_subquery_entity_links) == 0
        ):
            return LinkableElementFilter(
                element_names=element_names, without_any_of=frozenset({LinkableElementProperty.METRIC})
            )

        return LinkableElementFilter(
            element_names=element_names,
        )



from __future__ import annotations

from dataclasses import dataclass

from dbt_semantic_interfaces.naming.keywords import METRIC_TIME_ELEMENT_NAME
from dbt_semantic_interfaces.references import MetricReference
from typing_extensions import override

from metricflow_semantics.query.group_by_item.resolution_path import MetricFlowQueryResolutionPath
from metricflow_semantics.query.issues.issues_base import (
    MetricFlowQueryIssueType,
    MetricFlowQueryResolutionIssue,
)
from metricflow_semantics.query.resolver_inputs.base_resolver_inputs import MetricFlowQueryResolverInput


@dataclass(frozen=True)
class CumulativeMetricRequiresMetricTimeIssue(MetricFlowQueryResolutionIssue):
    """Describes an issue with a query that includes a cumulative metric but does not include metric_time."""

    metric_reference: MetricReference

    @override
    def ui_description(self, associated_input: MetricFlowQueryResolverInput) -> str:
        return (
            f"The query includes a cumulative metric {repr(self.metric_reference.element_name)} that does not "
            f"accumulate over all-time, but the group-by items do not include {repr(METRIC_TIME_ELEMENT_NAME)} "
            "or the metric's agg_time_dimension."
        )

    @override
    def with_path_prefix(self, path_prefix: MetricFlowQueryResolutionPath) -> CumulativeMetricRequiresMetricTimeIssue:
        return CumulativeMetricRequiresMetricTimeIssue(
            issue_type=self.issue_type,
            parent_issues=self.parent_issues,
            query_resolution_path=self.query_resolution_path.with_path_prefix(path_prefix),
            metric_reference=self.metric_reference,
        )

    @staticmethod
    def from_parameters(  # noqa: D102
        metric_reference: MetricReference, query_resolution_path: MetricFlowQueryResolutionPath
    ) -> CumulativeMetricRequiresMetricTimeIssue:
        return CumulativeMetricRequiresMetricTimeIssue(
            issue_type=MetricFlowQueryIssueType.ERROR,
            parent_issues=(),
            query_resolution_path=query_resolution_path,
            metric_reference=metric_reference,
        )



from __future__ import annotations

from dataclasses import dataclass
from typing import List, Optional, Sequence, Tuple

from dbt_semantic_interfaces.implementations.elements.dimension import PydanticDimensionTypeParams
from dbt_semantic_interfaces.protocols.dimension import (
    Dimension as SemanticManifestDimension,
)
from dbt_semantic_interfaces.protocols.dimension import (
    DimensionType,
    DimensionTypeParams,
)
from dbt_semantic_interfaces.protocols.entity import Entity as SemanticManifestEntity
from dbt_semantic_interfaces.protocols.export import Export
from dbt_semantic_interfaces.protocols.measure import MeasureAggregationParameters
from dbt_semantic_interfaces.protocols.metadata import Metadata
from dbt_semantic_interfaces.protocols.metric import Metric as SemanticManifestMetric
from dbt_semantic_interfaces.protocols.metric import MetricConfig, MetricInputMeasure, MetricType, MetricTypeParams
from dbt_semantic_interfaces.protocols.saved_query import (
    SavedQuery as SemanticManifestSavedQuery,
)
from dbt_semantic_interfaces.protocols.saved_query import (
    SavedQueryQueryParams,
)
from dbt_semantic_interfaces.protocols.where_filter import WhereFilterIntersection
from dbt_semantic_interfaces.references import EntityReference
from dbt_semantic_interfaces.transformations.add_input_metric_measures import AddInputMetricMeasuresRule
from dbt_semantic_interfaces.type_enums.aggregation_type import AggregationType
from dbt_semantic_interfaces.type_enums.entity_type import EntityType
from metricflow_semantics.naming.linkable_spec_name import StructuredLinkableSpecName
from metricflow_semantics.specs.dimension_spec import DimensionSpec


@dataclass(frozen=True)
class Metric:
    """Dataclass representation of a Metric."""

    name: str
    description: Optional[str]
    type: MetricType
    type_params: MetricTypeParams
    filter: Optional[WhereFilterIntersection]
    metadata: Optional[Metadata]
    dimensions: List[Dimension]
    label: Optional[str]
    config: Optional[MetricConfig]

    @classmethod
    def from_pydantic(cls, pydantic_metric: SemanticManifestMetric, dimensions: List[Dimension]) -> Metric:
        """Build from pydantic Metric object and list of Dimensions."""
        return cls(
            name=pydantic_metric.name,
            description=pydantic_metric.description,
            type=pydantic_metric.type,
            type_params=pydantic_metric.type_params,
            filter=pydantic_metric.filter,
            metadata=pydantic_metric.metadata,
            dimensions=dimensions,
            label=pydantic_metric.label,
            config=pydantic_metric.config,
        )

    @property
    def input_measures(self) -> Sequence[MetricInputMeasure]:
        """Return the complete list of input measure configurations for this metric."""
        assert self.type_params.input_measures, (
            f"Metric {self.name} should have had input_measures populated by "
            f"{AddInputMetricMeasuresRule.__class__.__name__}"
        )
        return self.type_params.input_measures


@dataclass(frozen=True)
class Dimension:
    """Dataclass representation of a Dimension."""

    name: str
    qualified_name: str
    description: Optional[str]
    type: DimensionType
    entity_links: Tuple[EntityReference, ...]
    type_params: Optional[DimensionTypeParams]
    metadata: Optional[Metadata]
    is_partition: bool = False
    expr: Optional[str] = None
    label: Optional[str] = None

    @classmethod
    def from_pydantic(
        cls,
        pydantic_dimension: SemanticManifestDimension,
        entity_links: Tuple[EntityReference, ...],
    ) -> Dimension:
        """Build from pydantic Dimension and entity_key."""
        qualified_name = DimensionSpec(element_name=pydantic_dimension.name, entity_links=entity_links).qualified_name
        parsed_type_params: Optional[DimensionTypeParams] = None
        if pydantic_dimension.type_params:
            parsed_type_params = PydanticDimensionTypeParams(
                time_granularity=pydantic_dimension.type_params.time_granularity,
                validity_params=pydantic_dimension.type_params.validity_params,
            )
        return cls(
            name=pydantic_dimension.name,
            qualified_name=qualified_name,
            description=pydantic_dimension.description,
            type=pydantic_dimension.type,
            type_params=parsed_type_params,
            metadata=pydantic_dimension.metadata,
            is_partition=pydantic_dimension.is_partition,
            expr=pydantic_dimension.expr,
            label=pydantic_dimension.label,
            entity_links=entity_links,
        )

    @property
    def granularity_free_qualified_name(self) -> str:
        """Renders the qualified name without the granularity suffix.

        In the list metrics and list dimensions outputs we want to render the qualified name of the dimension, but
        without including the base granularity for time dimensions. This method is useful in those contexts.

        Note: in most cases you should be using the qualified_name - this is only useful in cases where the
        Dimension set has de-duplicated TimeDimensions such that you never have more than one granularity
        in your set for each TimeDimension.
        """
        return StructuredLinkableSpecName(
            entity_link_names=tuple(e.element_name for e in self.entity_links), element_name=self.name
        ).qualified_name


@dataclass(frozen=True)
class Entity:
    """Dataclass representation of a Entity."""

    name: str
    description: Optional[str]
    type: EntityType
    role: Optional[str]
    expr: Optional[str] = None

    @classmethod
    def from_pydantic(cls, pydantic_entity: SemanticManifestEntity) -> Entity:
        """Build from pydantic Entity."""
        return cls(
            name=pydantic_entity.name,
            description=pydantic_entity.description,
            type=pydantic_entity.type,
            role=pydantic_entity.role,
            expr=pydantic_entity.expr,
        )


@dataclass(frozen=True)
class Measure:
    """Dataclass representation of a Measure."""

    name: str
    agg: AggregationType
    agg_time_dimension: str
    description: Optional[str] = None
    expr: Optional[str] = None
    agg_params: Optional[MeasureAggregationParameters] = None


@dataclass(frozen=True)
class SavedQuery:
    """Dataclass representation of a SavedQuery."""

    name: str
    description: Optional[str]
    label: Optional[str]
    query_params: SavedQueryQueryParams
    metadata: Optional[Metadata]
    exports: Sequence[Export]

    @classmethod
    def from_pydantic(cls, pydantic_saved_query: SemanticManifestSavedQuery) -> SavedQuery:
        """Build from pydantic SavedQuery object."""
        return cls(
            name=pydantic_saved_query.name,
            description=pydantic_saved_query.description,
            label=pydantic_saved_query.label,
            query_params=pydantic_saved_query.query_params,
            metadata=pydantic_saved_query.metadata,
            exports=pydantic_saved_query.exports,
        )



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional, Sequence, Tuple

from dbt_semantic_interfaces.protocols import MetricTimeWindow
from dbt_semantic_interfaces.type_enums import TimeGranularity
from metricflow_semantics.dag.id_prefix import IdPrefix, StaticIdPrefix
from metricflow_semantics.dag.mf_dag import DisplayedProperty
from metricflow_semantics.filters.time_constraint import TimeRangeConstraint
from metricflow_semantics.specs.time_dimension_spec import TimeDimensionSpec
from metricflow_semantics.visitor import VisitorOutputT

from metricflow.dataflow.dataflow_plan import DataflowPlanNode
from metricflow.dataflow.dataflow_plan_visitor import DataflowPlanNodeVisitor


@dataclass(frozen=True, eq=False)
class JoinOverTimeRangeNode(DataflowPlanNode):
    """A node that allows for cumulative metric computation by doing a self join across a cumulative date range.

    Attributes:
        queried_agg_time_dimension_specs: Time dimension specs that will be selected from the time spine table.
        window: Time window to join over.
        grain_to_date: Indicates time range should start from the beginning of this time granularity (e.g., month to day).
        time_range_constraint: Time range to aggregate over.
    """

    queried_agg_time_dimension_specs: Tuple[TimeDimensionSpec, ...]
    window: Optional[MetricTimeWindow]
    grain_to_date: Optional[TimeGranularity]
    time_range_constraint: Optional[TimeRangeConstraint]

    def __post_init__(self) -> None:  # noqa: D105
        super().__post_init__()
        assert len(self.parent_nodes) == 1

    @staticmethod
    def create(  # noqa: D102
        parent_node: DataflowPlanNode,
        queried_agg_time_dimension_specs: Tuple[TimeDimensionSpec, ...],
        window: Optional[MetricTimeWindow] = None,
        grain_to_date: Optional[TimeGranularity] = None,
        time_range_constraint: Optional[TimeRangeConstraint] = None,
    ) -> JoinOverTimeRangeNode:
        if window and grain_to_date:
            raise RuntimeError(
                f"This node cannot be initialized with both window and grain_to_date set. This configuration should "
                f"have been prevented by model validation. window: {window}. grain_to_date: {grain_to_date}."
            )
        return JoinOverTimeRangeNode(
            parent_nodes=(parent_node,),
            queried_agg_time_dimension_specs=queried_agg_time_dimension_specs,
            window=window,
            grain_to_date=grain_to_date,
            time_range_constraint=time_range_constraint,
        )

    @classmethod
    def id_prefix(cls) -> IdPrefix:  # noqa: D102
        return StaticIdPrefix.DATAFLOW_NODE_JOIN_SELF_OVER_TIME_RANGE_ID_PREFIX

    def accept(self, visitor: DataflowPlanNodeVisitor[VisitorOutputT]) -> VisitorOutputT:  # noqa: D102
        return visitor.visit_join_over_time_range_node(self)

    @property
    def description(self) -> str:  # noqa: D102
        return """Join Self Over Time Range"""

    @property
    def parent_node(self) -> DataflowPlanNode:  # noqa: D102
        return self.parent_nodes[0]

    @property
    def displayed_properties(self) -> Sequence[DisplayedProperty]:  # noqa: D102
        displayed_properties = tuple(super().displayed_properties)
        displayed_properties += (
            DisplayedProperty("queried_agg_time_dimension_specs", self.queried_agg_time_dimension_specs),
        )
        if self.window:
            displayed_properties += (DisplayedProperty("window", self.window),)
        if self.grain_to_date:
            displayed_properties += (DisplayedProperty("grain_to_date", self.grain_to_date),)
        if self.time_range_constraint:
            displayed_properties += (DisplayedProperty("time_range_constraint", self.time_range_constraint),)
        return displayed_properties

    def functionally_identical(self, other_node: DataflowPlanNode) -> bool:  # noqa: D102
        return (
            isinstance(other_node, self.__class__)
            and other_node.grain_to_date == self.grain_to_date
            and other_node.window == self.window
            and other_node.time_range_constraint == self.time_range_constraint
            and other_node.queried_agg_time_dimension_specs == self.queried_agg_time_dimension_specs
        )

    def with_new_parents(self, new_parent_nodes: Sequence[DataflowPlanNode]) -> JoinOverTimeRangeNode:  # noqa: D102
        assert len(new_parent_nodes) == 1
        return JoinOverTimeRangeNode.create(
            parent_node=new_parent_nodes[0],
            window=self.window,
            grain_to_date=self.grain_to_date,
            time_range_constraint=self.time_range_constraint,
            queried_agg_time_dimension_specs=self.queried_agg_time_dimension_specs,
        )



from __future__ import annotations

from dataclasses import dataclass
from typing import Generator

import pytest

from metricflow_semantics.dag.sequential_id import SequentialIdGenerator


@pytest.fixture(autouse=True, scope="function")
def setup_id_generators() -> Generator[None, None, None]:
    """Setup ID generation to start numbering at a specific value to get repeatability in generated IDs.

    Plan outputs contain IDs, so if the IDs are not consistent from run to run, there will be diffs in the actual vs.
    expected outputs during a test.

    Fixtures may generate IDs, so this needs to be done before every test.
    """
    with SequentialIdGenerator.id_number_space(start_value=IdNumberSpace.for_test_start().start_value):
        yield None


@dataclass(frozen=True)
class IdNumberSpace:
    """Defines the numbering of IDs when setting up tests and test fixtures."""

    start_value: int

    @staticmethod
    def for_test_start() -> IdNumberSpace:
        """Before each test run."""
        return IdNumberSpace(0)

    @staticmethod
    def for_block(block_index: int, block_size: int = 2000, first_block_offset: int = 10000) -> IdNumberSpace:
        """Used to define fixed-size blocks of ID number spaces.

        This is useful for creating unique / non-overlapping IDs for different groups of objects.
        """
        if not block_index >= 0:
            raise RuntimeError(f"block_index should be >= 0. Got: {block_index}")
        if not block_size > 1:
            raise RuntimeError(f"block_size should be > 1. Got: {block_size}")
        if not first_block_offset >= 0:
            raise RuntimeError(f"first_block_offset should be >= 0. Got: {first_block_offset}")
        return IdNumberSpace(first_block_offset + block_size * block_index)



from __future__ import annotations

import logging

from typing_extensions import override

from metricflow.sql.optimizer.sql_query_plan_optimizer import SqlPlanOptimizer
from metricflow.sql.sql_plan import (
    SqlCreateTableAsNode,
    SqlCteNode,
    SqlJoinDescription,
    SqlOrderByDescription,
    SqlPlanNode,
    SqlPlanNodeVisitor,
    SqlSelectColumn,
    SqlSelectQueryFromClauseNode,
    SqlSelectStatementNode,
    SqlTableNode,
)

logger = logging.getLogger(__name__)


class SqlTableAliasSimplifierVisitor(SqlPlanNodeVisitor[SqlPlanNode]):
    """Visits the SQL query plan to see if table aliases can be omitted when rendering column references."""

    @override
    def visit_cte_node(self, node: SqlCteNode) -> SqlPlanNode:
        return node.with_new_select(node.select_statement.accept(self))

    def visit_select_statement_node(self, node: SqlSelectStatementNode) -> SqlPlanNode:  # noqa: D102
        # If there is only a single source in the SELECT, no table aliases are required since there's no ambiguity.
        should_simplify_table_aliases = len(node.join_descs) == 0

        if should_simplify_table_aliases:
            return SqlSelectStatementNode.create(
                description=node.description,
                select_columns=tuple(
                    SqlSelectColumn(expr=x.expr.rewrite(should_render_table_alias=False), column_alias=x.column_alias)
                    for x in node.select_columns
                ),
                from_source=node.from_source.accept(self),
                from_source_alias=node.from_source_alias,
                cte_sources=tuple(
                    cte_source.with_new_select(cte_source.select_statement.accept(self))
                    for cte_source in node.cte_sources
                ),
                group_bys=tuple(
                    SqlSelectColumn(expr=x.expr.rewrite(should_render_table_alias=False), column_alias=x.column_alias)
                    for x in node.group_bys
                ),
                order_bys=tuple(
                    SqlOrderByDescription(expr=x.expr.rewrite(should_render_table_alias=False), desc=x.desc)
                    for x in node.order_bys
                ),
                where=node.where.rewrite(should_render_table_alias=False) if node.where else None,
                limit=node.limit,
                distinct=node.distinct,
            )

        return SqlSelectStatementNode.create(
            description=node.description,
            select_columns=node.select_columns,
            from_source=node.from_source.accept(self),
            from_source_alias=node.from_source_alias,
            cte_sources=tuple(
                cte_source.with_new_select(cte_source.select_statement.accept(self)) for cte_source in node.cte_sources
            ),
            join_descs=tuple(
                SqlJoinDescription(
                    right_source=x.right_source.accept(self),
                    right_source_alias=x.right_source_alias,
                    on_condition=x.on_condition,
                    join_type=x.join_type,
                )
                for x in node.join_descs
            ),
            group_bys=node.group_bys,
            order_bys=node.order_bys,
            where=node.where,
            limit=node.limit,
            distinct=node.distinct,
        )

    def visit_table_node(self, node: SqlTableNode) -> SqlPlanNode:  # noqa: D102
        return node

    def visit_query_from_clause_node(self, node: SqlSelectQueryFromClauseNode) -> SqlPlanNode:  # noqa: D102
        return node

    def visit_create_table_as_node(self, node: SqlCreateTableAsNode) -> SqlPlanNode:  # noqa: D102
        return SqlCreateTableAsNode.create(
            sql_table=node.sql_table,
            parent_node=node.parent_node.accept(self),
        )


class SqlTableAliasSimplifier(SqlPlanOptimizer):
    """Simplify queries by eliminating table aliases when possible.

    e.g. from

    SELECT b.foo
    FROM (
      SELECT a.foo FROM bar a
    ) b

    to

    SELECT foo
    FROM (
      SELECT foo FROM bar a
    ) b
    """

    def optimize(self, node: SqlPlanNode) -> SqlPlanNode:  # noqa: D102
        return node.accept(SqlTableAliasSimplifierVisitor())



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Sequence

from metricflow_semantics.model.semantics.element_filter import LinkableElementFilter

if TYPE_CHECKING:
    from metricflow_semantics.specs.instance_spec import InstanceSpec


class SpecPattern(ABC):
    """A pattern is used to select specs from a group of candidate specs based on class-defined criteria.

    This could be named SpecFilter as well, but a filter is often used in the context of the WhereFilter.
    """

    @abstractmethod
    def match(self, candidate_specs: Sequence[InstanceSpec]) -> Sequence[InstanceSpec]:
        """Given candidate specs, return the ones that match this pattern."""
        raise NotImplementedError

    def matches_any(self, candidate_specs: Sequence[InstanceSpec]) -> bool:
        """Returns true if this spec matches any of the given specs."""
        return len(self.match(candidate_specs)) > 0

    @property
    def element_pre_filter(self) -> LinkableElementFilter:
        """Returns a filter for a `LinkableElementSet` that can reduce the number of items to match.

        i.e. the filter can produce a superset of the elements that will match.
        """
        return LinkableElementFilter()



"""Functions to help generate a text representation of a DAG."""

from __future__ import annotations

import logging
import threading
import typing
from contextlib import contextmanager
from typing import Iterator, Optional

from metricflow_semantics.mf_logging.formatting import indent

if typing.TYPE_CHECKING:
    from metricflow_semantics.dag.mf_dag import DagNode, DagNodeT, DisplayedProperty, MetricFlowDag

from metricflow_semantics.mf_logging.pretty_print import mf_pformat

logger = logging.getLogger(__name__)


class MaxWidthTracker:
    """Helps to number columns remaining as the DAG is formatted to text recursively.

    This is needed as the parents of a given node are formatted with an indent. For example, if we want to print a DAG
    with a column width of 80 and the indent is 4 spaces, then we would want to print the parents with a column width of
    76.
    """

    def __init__(self, max_width: int) -> None:  # noqa: D107
        self._current_max_width = max_width

    @contextmanager
    def update_max_width_for_indented_section(self, indent_prefix: str) -> Iterator[None]:
        """Context manager used to wrap the code that prints an indented section."""
        previous_max_width = self._current_max_width
        self._current_max_width = max(1, self._current_max_width - len(indent_prefix))
        yield None
        self._current_max_width = previous_max_width

    @property
    def current_max_width(self) -> int:  # noqa: D102
        return self._current_max_width


class MetricFlowDagTextFormatter:
    """Converts the given node and parents (recursively) to a text representation.

    The text representation should be similar to the XML format where the parents are printed in an indented section.:

    <FilterElementsNode>
      ...
    </>
    """

    def __init__(
        self, max_width: int = 120, node_parent_indent_prefix: str = "    ", value_indent_prefix: str = "  "
    ) -> None:
        """Initializer.

        Args:
            max_width: Try to keep the generated text to this many columns wide.
            node_parent_indent_prefix: When printing a parent node, use this as a prefix for the indent.
            value_indent_prefix: When printing the values for a DisplayedProperty, use this indent.
        """
        self._max_width = max_width
        self._node_parent_indent_prefix = node_parent_indent_prefix
        self._value_indent_prefix = value_indent_prefix

        # In case this gets used in a multi-threaded context, use a thread-local variable since it has mutable state.
        self._thread_local_data = threading.local()

    @property
    def _max_width_tracker(self) -> MaxWidthTracker:
        if not hasattr(self._thread_local_data, "max_width_tracker"):
            self._thread_local_data.max_width_tracker = MaxWidthTracker(self._max_width)
        return self._thread_local_data.max_width_tracker

    def _displayed_property_on_one_line(self, displayed_property: DisplayedProperty) -> str:
        key = displayed_property.key
        value = mf_pformat(displayed_property.value, max_line_length=self._max_width_tracker.current_max_width)
        return f"<!-- {key} = {value} -->"

    def _format_to_text(self, node: DagNode, inner_contents: Optional[str]) -> str:
        """Convert the given node to the text representation.

        Properties is a dict from the property name to the property value that should be printed for the node.
        """
        # Generate the descriptions for the node
        node_fields = []
        max_width = self._max_width_tracker.current_max_width
        for displayed_property in node.displayed_properties:
            # See if the displayed property can be printed on one line.
            displayed_property_on_one_line = self._displayed_property_on_one_line(displayed_property)
            if len(displayed_property_on_one_line) <= max_width:
                node_fields.append(displayed_property_on_one_line)
                continue

            # If not, split them into multiple lines.
            value_str = mf_pformat(
                displayed_property.value,
                # The string representation of displayed_property.value will be wrapped with "<!-- ", " -->" so subtract
                # the width of those.
                max_line_length=max(1, max_width - len("<!-- ") - len(" -->")),
                indent_prefix=self._value_indent_prefix,
            )

            # Figure out the max width of the value so that we can add appropriate spacing so that the "<!--" / "-->"
            # line up.
            value_str_split = value_str.split("\n")
            max_value_str_length = max([len(x) for x in value_str_split])

            # Print the key on multiple lines.
            key = displayed_property.key
            # Add padding so that all fields of this object have <!-- and --> that align.
            key_padding = " " * (
                (len("<!-- ") + len(self._value_indent_prefix) + max_value_str_length + len(" -->"))
                - len("<!-- ")
                - len(key)
                - len(" = ")
                - len(" -->")
            )

            node_fields.append(f"<!-- {key} = {key_padding} -->")

            # Print the lines for the value in an indented section.
            for value_str in value_str_split:
                value_padding = " " * (
                    (len("<!-- ") + len(self._value_indent_prefix) + max_value_str_length + len(" -->"))
                    - len("<!-- ")
                    - len(self._value_indent_prefix)
                    - len(value_str)
                    - len(" -->")
                )
                node_fields.append(f"<!-- {self._value_indent_prefix}{value_str}{value_padding} -->")

        node_class = node.__class__.__name__
        if not node_fields and not inner_contents:
            return f"<{node_class}/>"

        lines = [f"<{node_class}>"]
        for line in node_fields:
            lines.append(indent(line, indent_prefix=self._node_parent_indent_prefix))
        if inner_contents:
            lines.append(indent(inner_contents, indent_prefix=self._node_parent_indent_prefix))
        lines.append(f"</{node_class}>")
        return "\n".join(lines)

    def _recursively_format_to_text(self, node: DagNode) -> str:
        """Converts the node and its parents to a text representation.

        The text representation is similar to XML.
        """
        parent_node_descriptions = []
        with self._max_width_tracker.update_max_width_for_indented_section(
            indent_prefix=self._node_parent_indent_prefix
        ):
            for parent_node in node.parent_nodes:
                parent_node_descriptions.append(self._recursively_format_to_text(parent_node))

        return self._format_to_text(node=node, inner_contents="\n".join(parent_node_descriptions))

    def dag_to_text(self, dag: MetricFlowDag[DagNodeT]) -> str:
        """Converts the DAG to a text representation that can be used for logging / tests.

        The text representation is similar to XML.
        """
        try:
            # Convert each of the components that are associated with the sink nodes to a text representation.
            component_from_sink_nodes_as_text = []

            with self._max_width_tracker.update_max_width_for_indented_section(self._node_parent_indent_prefix):
                for sink_node in dag.sink_nodes:
                    component_from_sink_nodes_as_text.append(self.dag_component_to_text(sink_node))

            # Under <DataflowPlan>, render all components.
            node_class = dag.__class__.__name__
            if len(component_from_sink_nodes_as_text) == 0:
                return f"<{node_class}/>"

            lines = [f"<{node_class}>"]
            for line in component_from_sink_nodes_as_text:
                lines.append(indent(line, indent_prefix=self._node_parent_indent_prefix))
            lines.append(f"</{node_class}>")

            return "\n".join(lines)

        except Exception:
            logger.exception(
                f"Got an exception while converting {dag} to text. This exception will be swallowed, and the built-in "
                f"string representation will be returned instead."
            )
            return str(dag)

    def dag_component_to_text(self, dag_component_leaf_node: DagNode) -> str:
        """Convert the DAG component starting from the given leaf node to a text representation."""
        try:
            return self._recursively_format_to_text(dag_component_leaf_node)
        except Exception:
            logger.exception(f"Got an exception while converting {dag_component_leaf_node} to text")
            return str(dag_component_leaf_node)



from __future__ import annotations

from dataclasses import dataclass
from typing import Sequence

from dbt_semantic_interfaces.references import MeasureReference, MetricReference
from typing_extensions import override

from metricflow_semantics.dag.id_prefix import IdPrefix, StaticIdPrefix
from metricflow_semantics.dag.mf_dag import DisplayedProperty
from metricflow_semantics.query.group_by_item.resolution_dag.resolution_nodes.base_node import (
    GroupByItemResolutionNode,
    GroupByItemResolutionNodeSet,
    GroupByItemResolutionNodeVisitor,
)
from metricflow_semantics.visitor import VisitorOutputT


@dataclass(frozen=True)
class MeasureGroupByItemSourceNode(GroupByItemResolutionNode):
    """Outputs group-by-items for a measure.

    Attributes:
        measure_reference: Get the group-by items for this measure.
        child_metric_reference: The metric that uses this measure.
    """

    measure_reference: MeasureReference
    child_metric_reference: MetricReference

    def __post_init__(self) -> None:  # noqa: D105
        super().__post_init__()
        assert len(self.parent_nodes) == 0

    @staticmethod
    def create(  # noqa: D102
        measure_reference: MeasureReference,
        child_metric_reference: MetricReference,
    ) -> MeasureGroupByItemSourceNode:
        return MeasureGroupByItemSourceNode(
            parent_nodes=(),
            measure_reference=measure_reference,
            child_metric_reference=child_metric_reference,
        )

    @override
    def accept(self, visitor: GroupByItemResolutionNodeVisitor[VisitorOutputT]) -> VisitorOutputT:
        return visitor.visit_measure_node(self)

    @property
    @override
    def description(self) -> str:
        return "Output group-by-items available for this measure."

    @classmethod
    @override
    def id_prefix(cls) -> IdPrefix:
        return StaticIdPrefix.MEASURE_GROUP_BY_ITEM_RESOLUTION_NODE

    @property
    @override
    def displayed_properties(self) -> Sequence[DisplayedProperty]:
        return tuple(super().displayed_properties) + (
            DisplayedProperty(
                key="measure_reference",
                value=str(self.measure_reference),
            ),
            DisplayedProperty(
                key="child_metric_reference",
                value=str(self.child_metric_reference),
            ),
        )

    @property
    @override
    def ui_description(self) -> str:
        return f"Measure({repr(self.measure_reference.element_name)})"

    @override
    def _self_set(self) -> GroupByItemResolutionNodeSet:
        return GroupByItemResolutionNodeSet(measure_nodes=(self,))



from __future__ import annotations

from dataclasses import dataclass
from typing import List, Sequence

from typing_extensions import override

from metricflow_semantics.model.linkable_element_property import LinkableElementProperty
from metricflow_semantics.model.semantics.element_filter import LinkableElementFilter
from metricflow_semantics.specs.instance_spec import InstanceSpec, LinkableInstanceSpec
from metricflow_semantics.specs.patterns.spec_pattern import SpecPattern
from metricflow_semantics.specs.spec_set import group_specs_by_type


@dataclass(frozen=True)
class NoGroupByMetricPattern(SpecPattern):
    """Matches to linkable specs, but only if they're not group by metrics.

    Group by metrics are allowed in filters but not in the query input group by.
    """

    @override
    def match(self, candidate_specs: Sequence[InstanceSpec]) -> Sequence[LinkableInstanceSpec]:
        specs_to_return: List[LinkableInstanceSpec] = []
        spec_set = group_specs_by_type(candidate_specs)
        specs_to_return.extend(spec_set.time_dimension_specs)
        specs_to_return.extend(spec_set.dimension_specs)
        specs_to_return.extend(spec_set.entity_specs)

        return specs_to_return

    @property
    @override
    def element_pre_filter(self) -> LinkableElementFilter:
        return LinkableElementFilter(without_any_of=frozenset({LinkableElementProperty.METRIC}))



from __future__ import annotations

import logging

from metricflow_semantics.mf_logging.lazy_formattable import LazyFormat
from typing_extensions import override

from metricflow.sql.optimizer.cte_mapping_lookup_builder import SqlCteAliasMappingLookupBuilderVisitor
from metricflow.sql.optimizer.required_column_aliases import SqlMapRequiredColumnAliasesVisitor
from metricflow.sql.optimizer.sql_query_plan_optimizer import SqlPlanOptimizer
from metricflow.sql.optimizer.tag_column_aliases import NodeToColumnAliasMapping
from metricflow.sql.sql_plan import (
    SqlCreateTableAsNode,
    SqlCteAliasMapping,
    SqlCteNode,
    SqlPlanNode,
    SqlPlanNodeVisitor,
    SqlSelectQueryFromClauseNode,
    SqlSelectStatementNode,
    SqlTableNode,
)

logger = logging.getLogger(__name__)


class SqlColumnPrunerVisitor(SqlPlanNodeVisitor[SqlPlanNode]):
    """Removes unnecessary columns from SELECT statements in the SQL query plan.

    This requires a set of tagged column aliases that should be kept for each SQL node.
    """

    def __init__(
        self,
        required_alias_mapping: NodeToColumnAliasMapping,
    ) -> None:
        """Constructor.

        Args:
            required_alias_mapping: Describes columns aliases that should be kept / not pruned for each node.
        """
        self._required_alias_mapping = required_alias_mapping

    def visit_select_statement_node(self, node: SqlSelectStatementNode) -> SqlPlanNode:  # noqa: D102
        # Remove columns that are not needed from this SELECT statement because the parent SELECT statement doesn't
        # need them. However, keep columns that are in group bys because that changes the meaning of the query.
        # Similarly, if this node is a distinct select node, keep all columns as it may return a different result set.
        required_column_aliases = self._required_alias_mapping.get_aliases(node)
        if required_column_aliases is None:
            logger.error(
                f"Did not find {node.node_id=} in the required alias mapping. Returning the non-pruned version "
                f"as it should be valid SQL, but this is a bug and should be investigated."
            )
            return node

        if len(required_column_aliases) == 0:
            logger.error(
                f"Got no required column aliases for {node}. Returning the non-pruned version as it should be valid "
                f"SQL, but this is a bug and should be investigated."
            )
            return node

        pruned_select_columns = tuple(
            select_column
            for select_column in node.select_columns
            if select_column.column_alias in required_column_aliases
        )

        return SqlSelectStatementNode.create(
            description=node.description,
            select_columns=pruned_select_columns,
            from_source=node.from_source.accept(self),
            from_source_alias=node.from_source_alias,
            cte_sources=tuple(
                cte_source.with_new_select(cte_source.select_statement.accept(self)) for cte_source in node.cte_sources
            ),
            join_descs=tuple(
                join_desc.with_right_source(join_desc.right_source.accept(self)) for join_desc in node.join_descs
            ),
            group_bys=node.group_bys,
            order_bys=node.order_bys,
            where=node.where,
            limit=node.limit,
            distinct=node.distinct,
        )

    def visit_table_node(self, node: SqlTableNode) -> SqlPlanNode:
        """There are no SELECT columns in this node, so pruning cannot apply."""
        return node

    def visit_query_from_clause_node(self, node: SqlSelectQueryFromClauseNode) -> SqlPlanNode:
        """Pruning cannot be done here since this is an arbitrary user-provided SQL query."""
        return node

    def visit_create_table_as_node(self, node: SqlCreateTableAsNode) -> SqlPlanNode:  # noqa: D102
        return SqlCreateTableAsNode.create(
            sql_table=node.sql_table,
            parent_node=node.parent_node.accept(self),
        )

    @override
    def visit_cte_node(self, node: SqlCteNode) -> SqlPlanNode:
        return node.with_new_select(node.select_statement.accept(self))


class SqlColumnPrunerOptimizer(SqlPlanOptimizer):
    """Removes unnecessary columns in the SELECT statements."""

    def optimize(self, node: SqlPlanNode) -> SqlPlanNode:  # noqa: D102
        # ALl columns in the nearest SELECT node need to be kept as otherwise, the meaning of the query changes.
        required_select_columns = node.nearest_select_columns(SqlCteAliasMapping())

        # Can't prune without knowing the structure of the query.
        if required_select_columns is None:
            logger.error(
                LazyFormat(
                    "The columns required at this node can't be determined, so skipping column pruning",
                    node=node.structure_text(),
                    required_select_columns=required_select_columns,
                )
            )
            return node

        cte_alias_mapping_builder = SqlCteAliasMappingLookupBuilderVisitor()
        node.accept(cte_alias_mapping_builder)
        cte_alias_mapping_lookup = cte_alias_mapping_builder.cte_alias_mapping_lookup

        map_required_column_aliases_visitor = SqlMapRequiredColumnAliasesVisitor(
            start_node=node,
            cte_alias_mapping_lookup=cte_alias_mapping_lookup,
            required_column_aliases_in_start_node=frozenset(
                [select_column.column_alias for select_column in required_select_columns]
            ),
        )
        node.accept(map_required_column_aliases_visitor)

        # Re-write the query, removing unnecessary columns in the SELECT statements.
        pruning_visitor = SqlColumnPrunerVisitor(map_required_column_aliases_visitor.required_column_alias_mapping)
        return node.accept(pruning_visitor)



from __future__ import annotations

from dataclasses import dataclass
from typing import List, Sequence, Tuple, Union

from dbt_semantic_interfaces.protocols import WhereFilterIntersection
from dbt_semantic_interfaces.references import MetricReference
from typing_extensions import override

from metricflow_semantics.dag.id_prefix import IdPrefix, StaticIdPrefix
from metricflow_semantics.dag.mf_dag import DisplayedProperty
from metricflow_semantics.query.group_by_item.resolution_dag.resolution_nodes.base_node import (
    GroupByItemResolutionNode,
    GroupByItemResolutionNodeSet,
    GroupByItemResolutionNodeVisitor,
)
from metricflow_semantics.query.group_by_item.resolution_dag.resolution_nodes.metric_resolution_node import (
    MetricGroupByItemResolutionNode,
)
from metricflow_semantics.query.group_by_item.resolution_dag.resolution_nodes.no_metrics_query_source_node import (
    NoMetricsGroupByItemSourceNode,
)
from metricflow_semantics.visitor import VisitorOutputT


@dataclass(frozen=True)
class QueryGroupByItemResolutionNode(GroupByItemResolutionNode):
    """Output the group-by-items relevant to the query and based on the inputs.

    Attributes:
        parent_nodes: The parent nodes of this query.
        metrics_in_query: The metrics that are queried in this query.
        where_filter_intersection: The intersection of where filters.
    """

    parent_nodes: Tuple[Union[MetricGroupByItemResolutionNode, NoMetricsGroupByItemSourceNode], ...]
    metrics_in_query: Tuple[MetricReference, ...]
    where_filter_intersection: WhereFilterIntersection

    @staticmethod
    def create(  # noqa: D102
        parent_nodes: Sequence[Union[MetricGroupByItemResolutionNode, NoMetricsGroupByItemSourceNode]],
        metrics_in_query: Sequence[MetricReference],
        where_filter_intersection: WhereFilterIntersection,
    ) -> QueryGroupByItemResolutionNode:
        return QueryGroupByItemResolutionNode(
            parent_nodes=tuple(parent_nodes),
            metrics_in_query=tuple(metrics_in_query),
            where_filter_intersection=where_filter_intersection,
        )

    @override
    def accept(self, visitor: GroupByItemResolutionNodeVisitor[VisitorOutputT]) -> VisitorOutputT:
        return visitor.visit_query_node(self)

    @property
    @override
    def description(self) -> str:
        return "Output the group-by items for query."

    @classmethod
    @override
    def id_prefix(cls) -> IdPrefix:
        return StaticIdPrefix.QUERY_GROUP_BY_ITEM_RESOLUTION_NODE

    @property
    @override
    def displayed_properties(self) -> List[DisplayedProperty]:
        properties = list(super().displayed_properties)

        if len(self.metrics_in_query) > 0:
            properties.append(
                DisplayedProperty(
                    key="metrics_in_query",
                    value=[metric_reference.element_name for metric_reference in self.metrics_in_query],
                )
            )

        if len(self.where_filter_intersection.where_filters) > 0:
            properties.append(
                DisplayedProperty(
                    key="where_filter",
                    value=[
                        where_filter.where_sql_template for where_filter in self.where_filter_intersection.where_filters
                    ],
                )
            )

        return properties

    @property
    @override
    def ui_description(self) -> str:
        return f"Query({repr([metric_reference.element_name for metric_reference in self.metrics_in_query])})"

    @override
    def _self_set(self) -> GroupByItemResolutionNodeSet:
        return GroupByItemResolutionNodeSet(query_nodes=(self,))



from __future__ import annotations

from abc import ABC
from dataclasses import dataclass
from typing import Sequence

from metricflow_semantics.dag.id_prefix import IdPrefix, StaticIdPrefix
from metricflow_semantics.dag.mf_dag import DisplayedProperty
from metricflow_semantics.specs.time_dimension_spec import TimeDimensionSpec
from metricflow_semantics.visitor import VisitorOutputT

from metricflow.dataflow.dataflow_plan import DataflowPlanNode
from metricflow.dataflow.dataflow_plan_visitor import DataflowPlanNodeVisitor


@dataclass(frozen=True, eq=False)
class JoinToCustomGranularityNode(DataflowPlanNode, ABC):
    """Join parent dataset to time spine dataset to convert time dimension to a custom granularity.

    Args:
        time_dimension_spec: The time dimension spec with a custom granularity that will be satisfied by this node.
    """

    time_dimension_spec: TimeDimensionSpec

    def __post_init__(self) -> None:  # noqa: D105
        super().__post_init__()
        assert (
            self.time_dimension_spec.time_granularity.is_custom_granularity
        ), "Time granularity for time dimension spec in JoinToCustomGranularityNode must be qualified as custom granularity."
        f" Instead, found {self.time_dimension_spec.time_granularity.name}. This indicates internal misconfiguration."

    @staticmethod
    def create(  # noqa: D102
        parent_node: DataflowPlanNode, time_dimension_spec: TimeDimensionSpec
    ) -> JoinToCustomGranularityNode:
        return JoinToCustomGranularityNode(parent_nodes=(parent_node,), time_dimension_spec=time_dimension_spec)

    @classmethod
    def id_prefix(cls) -> IdPrefix:  # noqa: D102
        return StaticIdPrefix.DATAFLOW_NODE_JOIN_TO_CUSTOM_GRANULARITY_ID_PREFIX

    def accept(self, visitor: DataflowPlanNodeVisitor[VisitorOutputT]) -> VisitorOutputT:  # noqa: D102
        return visitor.visit_join_to_custom_granularity_node(self)

    @property
    def description(self) -> str:  # noqa: D102
        return """Join to Custom Granularity Dataset"""

    @property
    def displayed_properties(self) -> Sequence[DisplayedProperty]:  # noqa: D102
        return tuple(super().displayed_properties) + (
            DisplayedProperty("time_dimension_spec", self.time_dimension_spec),
        )

    @property
    def parent_node(self) -> DataflowPlanNode:  # noqa: D102
        return self.parent_nodes[0]

    def functionally_identical(self, other_node: DataflowPlanNode) -> bool:  # noqa: D102
        return isinstance(other_node, self.__class__) and other_node.time_dimension_spec == self.time_dimension_spec

    def with_new_parents(  # noqa: D102
        self, new_parent_nodes: Sequence[DataflowPlanNode]
    ) -> JoinToCustomGranularityNode:
        assert len(new_parent_nodes) == 1, "JoinToCustomGranularity accepts exactly one parent node."
        return JoinToCustomGranularityNode.create(
            parent_node=new_parent_nodes[0],
            time_dimension_spec=self.time_dimension_spec,
        )



from __future__ import annotations

from dataclasses import dataclass

from typing_extensions import override

from metricflow_semantics.query.group_by_item.resolution_path import MetricFlowQueryResolutionPath
from metricflow_semantics.query.issues.issues_base import (
    MetricFlowQueryIssueType,
    MetricFlowQueryResolutionIssue,
)
from metricflow_semantics.query.resolver_inputs.base_resolver_inputs import MetricFlowQueryResolverInput


@dataclass(frozen=True)
class InvalidLimitIssue(MetricFlowQueryResolutionIssue):
    """Describes an issue with the query where the limit is invalid."""

    limit: int

    @staticmethod
    def from_parameters(  # noqa: D102
        limit: int, query_resolution_path: MetricFlowQueryResolutionPath
    ) -> InvalidLimitIssue:  # noqa: D102
        return InvalidLimitIssue(
            issue_type=MetricFlowQueryIssueType.ERROR,
            parent_issues=(),
            query_resolution_path=query_resolution_path,
            limit=limit,
        )

    @override
    def ui_description(self, associated_input: MetricFlowQueryResolverInput) -> str:
        return f"The limit {repr(self.limit)} is not >= 0."

    @override
    def with_path_prefix(self, path_prefix: MetricFlowQueryResolutionPath) -> InvalidLimitIssue:
        return InvalidLimitIssue(
            issue_type=self.issue_type,
            parent_issues=tuple(issue.with_path_prefix(path_prefix) for issue in self.parent_issues),
            query_resolution_path=self.query_resolution_path.with_path_prefix(path_prefix),
            limit=self.limit,
        )



from __future__ import annotations

import logging
from dataclasses import dataclass
from hashlib import sha1
from typing import Any, Sequence, Tuple

from dbt_semantic_interfaces.dataclass_serialization import SerializableDataclass
from dbt_semantic_interfaces.type_enums import AggregationType, TimeGranularity

from metricflow_semantics.mf_logging.lazy_formattable import LazyFormat
from metricflow_semantics.naming.linkable_spec_name import DUNDER
from metricflow_semantics.specs.entity_spec import LinklessEntitySpec
from metricflow_semantics.specs.instance_spec import LinkableInstanceSpec
from metricflow_semantics.specs.time_dimension_spec import TimeDimensionSpec
from metricflow_semantics.sql.sql_column_type import SqlColumnType
from metricflow_semantics.time.granularity import ExpandedTimeGranularity

logger = logging.getLogger(__file__)


def hash_items(items: Sequence[SqlColumnType]) -> str:
    """Produces a hash from a list of strings."""
    hash_builder = sha1()
    for item in items:
        hash_builder.update(str(item).encode("utf-8"))
    return hash_builder.hexdigest()


@dataclass(frozen=True)
class NonAdditiveDimensionSpec(SerializableDataclass):
    """Spec representing non-additive dimension parameters for use within a MeasureSpec.

    This is sourced from the NonAdditiveDimensionParameters model object, which provides the parsed parameter set,
    while the spec contains the information needed for dataflow plan operations
    """

    name: str
    window_choice: AggregationType
    window_groupings: Tuple[str, ...] = ()

    def __post_init__(self) -> None:
        """Post init validator to ensure names with double-underscores are not allowed."""
        # TODO: [custom granularity] change this to an assertion once we're sure there aren't exceptions
        if not self.name.find(DUNDER) == -1:
            logger.warning(
                LazyFormat(
                    lambda: f"Non-additive dimension spec references a dimension name `{self.name}`, with added annotations, but it "
                    "should be a simple element name reference. This should have been blocked by model validation!"
                )
            )

    @property
    def bucket_hash(self) -> str:
        """Returns the hash value used for grouping equivalent params."""
        values = [self.window_choice.name, self.name]
        values.extend(sorted(self.window_groupings))
        return hash_items(values)

    def linkable_specs(self, non_additive_dimension_grain: TimeGranularity) -> Sequence[LinkableInstanceSpec]:
        """Return the set of linkable specs referenced by the NonAdditiveDimensionSpec.

        In practice, the name will always point to a time dimension. This method requires the time granularity
        provided in the model Dimension definition, which is why the input is typed as an enum value rather than
        an expanded granularity object - custom granularities are not eligible for consideration here.
        """
        return (
            TimeDimensionSpec(
                element_name=self.name,
                entity_links=(),
                time_granularity=ExpandedTimeGranularity.from_time_granularity(non_additive_dimension_grain),
            ),
        ) + tuple(LinklessEntitySpec.from_element_name(entity_name) for entity_name in self.window_groupings)

    def __eq__(self, other: Any) -> bool:  # type: ignore[misc]  # noqa: D105
        if not isinstance(other, NonAdditiveDimensionSpec):
            return False
        return self.bucket_hash == other.bucket_hash



from __future__ import annotations

from abc import abstractmethod
from enum import Enum
from typing import Protocol, Set

from dbt_semantic_interfaces.enum_extension import assert_values_exhausted
from dbt_semantic_interfaces.type_enums.time_granularity import TimeGranularity
from metricflow_semantics.sql.sql_bind_parameters import SqlBindParameterSet

from metricflow.data_table.mf_table import MetricFlowDataTable
from metricflow.sql.render.sql_plan_renderer import SqlPlanRenderer


class SqlEngine(Enum):
    """Enumeration of supported SQL engines.

    Values are normalized engine names used for things like snapshot file locations.
    """

    BIGQUERY = "BigQuery"
    DUCKDB = "DuckDB"
    REDSHIFT = "Redshift"
    POSTGRES = "Postgres"
    SNOWFLAKE = "Snowflake"
    DATABRICKS = "Databricks"
    TRINO = "Trino"

    @property
    def unsupported_granularities(self) -> Set[TimeGranularity]:
        """Granularities that can't be used with this SqlEngine.

        We allow the smallest granularity the SqlEngine supports for its base TIMESTAMP type and all our required
        operations (e.g., DATE_TRUNC). For example, when we added support for these granularities
        Trino supported more precise types for storage and access, but Trino's base TIMESTAMP type and
        DATE_TRUNC function only supported millisecond precision.
        """
        if self is SqlEngine.SNOWFLAKE:
            return set()
        elif self is SqlEngine.BIGQUERY:
            return {TimeGranularity.NANOSECOND}
        elif self is SqlEngine.DATABRICKS:
            return {TimeGranularity.NANOSECOND}
        elif self is SqlEngine.DUCKDB:
            return {TimeGranularity.NANOSECOND}
        elif self is SqlEngine.POSTGRES:
            return {TimeGranularity.NANOSECOND}
        elif self is SqlEngine.REDSHIFT:
            return {TimeGranularity.NANOSECOND}
        elif self is SqlEngine.TRINO:
            return {TimeGranularity.NANOSECOND, TimeGranularity.MICROSECOND}
        else:
            assert_values_exhausted(self)


class SqlClient(Protocol):
    """Base interface for SqlClient instances used inside MetricFlow.

    This provides the methods needed to execute SQL queries against the relevant Data Warehouse configuration.
    """

    @property
    @abstractmethod
    def sql_engine_type(self) -> SqlEngine:
        """Enumerated value representing the underlying SqlEngine for this SqlClient instance."""
        raise NotImplementedError

    @property
    @abstractmethod
    def sql_plan_renderer(self) -> SqlPlanRenderer:
        """Dialect-specific SQL plan renderer used for converting MetricFlow's query plan to executable SQL.

        This is bundled with the SqlClient partly as a convenenience for accessing a single instance of the renderer,
        and partly due to the close relationship between dialect and engine capabilities.
        """
        raise NotImplementedError

    @abstractmethod
    def query(
        self,
        stmt: str,
        sql_bind_parameter_set: SqlBindParameterSet = SqlBindParameterSet(),
    ) -> MetricFlowDataTable:
        """Base query method, upon execution will run a query that returns a pandas DataTable."""
        raise NotImplementedError

    @abstractmethod
    def execute(
        self,
        stmt: str,
        sql_bind_parameter_set: SqlBindParameterSet = SqlBindParameterSet(),
    ) -> None:
        """Base execute method."""
        raise NotImplementedError

    @abstractmethod
    def dry_run(
        self,
        stmt: str,
        sql_bind_parameter_set: SqlBindParameterSet = SqlBindParameterSet(),
    ) -> None:
        """Base dry_run method."""
        raise NotImplementedError

    @abstractmethod
    def close(self) -> None:
        """Close the connections / engines used by this client."""
        raise NotImplementedError

    @abstractmethod
    def render_bind_parameter_key(self, bind_parameter_key: str) -> str:
        """Wrap the bind parameter key with syntax accepted by engine."""
        raise NotImplementedError



from __future__ import annotations

from dataclasses import dataclass
from typing import Sequence

import rapidfuzz.fuzz
import rapidfuzz.process


@dataclass(frozen=True)
class ScoredItem:  # noqa: D101
    item_str: str
    # fuzz scores from 0..100, and the higher the score, the better the match.
    score: float


def top_fuzzy_matches(
    item: str,
    candidate_items: Sequence[str],
    max_matches: int = 6,
) -> Sequence[ScoredItem]:
    """Return the top items (by edit distance) in candidate_items that fuzzy matches the given item.

    Return scores from -1 -> 0 inclusive.
    """
    # In the case of a tie in score, items will be returned in the order they were passed in.
    # Sort candidate item inputs first for consistent results.
    sorted_candidate_items = sorted(candidate_items)

    scored_items = []

    # Rank choices by edit distance score.
    # extract() returns a tuple like (name, score)
    top_ranked_suggestions = sorted(
        rapidfuzz.process.extract(
            # This scorer seems to return the best results.
            item,
            sorted_candidate_items,
            limit=max_matches,
            scorer=rapidfuzz.fuzz.token_set_ratio,
        ),
        # Put the highest scoring item at the top of the list.
        key=lambda item_and_score_tuple: item_and_score_tuple[1],
        reverse=True,
    )

    for fuzz_tuple in top_ranked_suggestions:
        value = fuzz_tuple[0]
        score = fuzz_tuple[1]

        scored_items.append(ScoredItem(item_str=value, score=score))

    return scored_items
